#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Module to manage diverse mass-radius models.

This module provides tools for estimating the mass (or radius) from the radius
(or mass), using diverse power law models.
"""

# ============================================================================
# IMPORTS
# ============================================================================

import warnings
from typing import Tuple, Union

import numpy as np

from resokit.units import convert

# =============================================================================
# FUNCTIONS
# =============================================================================


def power_law(x: float, c: float, s: float, x0: float = 1.0) -> float:
    r"""Calculate a power-law.

    Equation: :math:`y = c \times \\left(\frac{x}{x}\\right)^s`

    Parameters
    ----------
    x : float
        Value to calculate the power-law.
    c : float
        Constant of the power-law relation.
    s : float
        Slope of the power-law relation.
    x0 : float, optional
        Reference value of the power-law.
        Default is 1.0.

    Returns
    -------
    float
        Result of the power-law.
    """
    return c * (x / x0) ** s


def power_law_error(
    x: float,
    x_err: float,
    c: float,
    c_err: float,
    s: float,
    s_err: float,
    x0: float = 1.0,
    x0_err: float = 0.0,
    y: float = 0,
) -> float:
    """Calculate the (naive propagation) error of a power-law.

    Equation: y_err = sqrt(
        (y / c  * c_err)^2
        + (y * log(x / x0) * s_err)^2
        + (y * s / x * x_err)^2
        + (y * s / x0 * x0_err)^2
    )

    Parameters
    ----------
    x : float
        Value to calculate the power-law.
    x_err : float
        Error of the value.
    c : float
        Constant of the power-law relation.
    c_err : float
        Error of the constant.
    s : float
        Slope of the power-law relation.
    s_err : float
        Error of the slope.
    x0 : float, optional. Default: 1
        Reference value of the power-law.
    x0_err : float, optional. Default: 0
        Error of the reference value.
    y : float, optional. Default: 0
        Value of the power-law.

    Returns
    -------
    float
        Error of the power-law.
    """
    if y == 0:
        y = power_law(x, c, s, x0)

    return np.sqrt(
        (y / c * c_err) ** 2  # dy/dc
        + (y * np.log(x / x0) * s_err) ** 2  # dy/ds
        + (y * s / x * x_err) ** 2  # dy/dx
        + (y * s / x0 * x0_err) ** 2  # dy/dx0
    )


def chen_kipp_2017_radius(mass: float) -> Tuple[float, tuple, tuple, tuple]:
    r"""Calculate the radius of a planet using the Chen & Kipping (2017).

    Power law approximation:
        :math:`radius = C \times mass^S`
    Citation:
        Chen, J., & Kipping, D. 2017, ApJ, 834, 17
    For a complete implementation of the method, see:
        https://github.com/chenjj2/forecaster

    Parameters
    ----------
    mass : float
        Mass of the planet, in Earth masses.

    Returns
    -------
    radius : float
        Radius of the planet, in Earth radii.
    c : tuple
        Constant of the power-law relation and its error.
    s : tuple
        Slope of the power-law relation and its error.
    x0 : tuple
        Reference value of the power-law and its error.
    """
    # Constants
    c1 = (1.008, 0.0046)
    c2 = (0.808119, 0.0172397)
    c3 = (17.738384, 5.851034)
    c4 = (0.00143, 0.000669)
    # Slopes
    s1 = (0.279, 0.0094)
    s2 = (0.589, 0.044)
    s3 = (-0.044, 0.019)
    s4 = (0.881, 0.025)
    # Reference value
    x0 = (1.0, 0.0)
    # Transition mass
    m1_tr = 2.04
    m2_tr = convert(
        0.414, from_units="mj", to_units="me"
    )  # 0.414 Jupiter masses
    m3_tr = convert(0.08, from_units="ms", to_units="me")  # 0.08 Solar masses

    if mass < m1_tr:  # First branch
        return power_law(mass, c1[0], s1[0], x0[0]), c1, s1, x0
    elif mass < m2_tr:  # Second branch
        return power_law(mass, c2[0], s2[0], x0[0]), c2, s2, x0
    elif mass < m3_tr:  # Third branch
        return power_law(mass, c3[0], s3[0], x0[0]), c3, s3, x0
    # Fourth branch
    return power_law(mass, c4[0], s4[0], x0[0]), c4, s4, x0


def chen_kipp_2017_mass(
    radius: float, trivariate: tuple = (0.15, 0.8), silent: bool = False
) -> Tuple[float, tuple, tuple, tuple]:
    r"""Calculate the mass of a planet using the Chen & Kipping (2017).

    Power law approximation:
        :math:`mass = x_0 \frac{1}{C} \times radius^{1/S}`
    Citation:
        Chen, J., & Kipping, D. 2017, ApJ, 834, 17
    For a complete implementation of the method, see:
        https://github.com/chenjj2/forecaster

    Parameters
    ----------
    radius : float
        Radius of the planet, in Earth radii.
    trivariate : tuple, optional. Default: (0.15, 0.8)
        Probabilities (from 0 to 1) that the returned radius that falls in the
        trivariate region is calculated with the second (left), ant then third
        (center) branch of the power-law approximation. The probability of
        using the fourth (right) branch is equal to 1 - sum(bivariate), so
        the sum of them must be lower equal than 1.
    silent : bool, optional. Default: False
        Whether to silence the warning if the radius falls in a
        multivariate region.

    Returns
    -------
    mass : float
        Mass of the planet, in Earth masses.
    inv_c : tuple
        Inverse of the constant of the power-law relation, and its error.
    inv_s : tuple
        Inverse of the slope of the power-law relation, and its error.
    x0 : tuple
        Reference value of the power-law, and its error.
    """
    # Constants
    c1 = (1.008, 0.0046)
    c2 = (0.808119, 0.0172397)
    c3 = (17.738384, 5.851034)
    c4 = (0.00143, 0.000669)
    # Slopes
    s1 = (0.279, 0.0094)
    s2 = (0.589, 0.044)
    s3 = (-0.044, 0.019)
    s4 = (0.881, 0.025)
    # Reference value
    x0 = (1.0, 0.0)
    # Transition radius
    r1_tr = (1.229836, 0.111458)
    r2_tr = (14.31101, 4.529131)
    r3_tr = (11.328892, 4.333345)

    # We use the inverse of the constant and slope,
    # so the error is recaclulated with propagation error.

    if radius < r1_tr[0]:  # First branch
        return (
            power_law(radius, x0[0], 1.0 / s1[0], c1[0]),
            x0,
            (1.0 / s1[0], s1[1] / s1[0]),
            c1,
        )
    elif radius > r2_tr[0]:  # Pure fourth branch
        return (
            power_law(radius, x0[0], 1.0 / s4[0], c4[0]),
            x0,
            (1.0 / s4[0], s4[1] / s4[0]),
            c4,
        )
    elif radius < r3_tr[0]:  # Pure second branch
        return (
            power_law(radius, x0[0], 1.0 / s2[0], c2[0]),
            x0,
            (1.0 / s2[0], s2[1] / s2[0]),
            c2,
        )

    # Trivariate region
    if not silent:
        warnings.warn(
            "Radius falls in the trivariate region: "
            + f"{r3_tr[0]} < R < {r2_tr[0]}"
            + "\n The mass-radius relation may not be accurate.",
            stacklevel=2,
        )

    if isinstance(trivariate, (float, int)) or len(trivariate) != 2:
        raise ValueError("Trivariate must be a tuple|list with length 2.")

    sumb = sum(trivariate)  # Probability of second branch
    if sumb < 0 or sumb > 1:
        raise ValueError("Sum of trivariate must be a number between 0 and 1.")

    prob = np.random.rand()  # Get a random probability

    if prob < trivariate[0]:  # Use second branch
        return (
            power_law(radius, x0[0], 1.0 / s2[0], c2[0]),
            x0,
            (1.0 / s2[0], s2[1] / s2[0]),
            c2,
        )
    elif prob < sumb:  # Use third branch
        return (
            power_law(radius, x0[0], 1.0 / s3[0], c3[0]),
            x0,
            (1.0 / s3[0], s3[1] / s3[0]),
            c3,
        )
    # Use fourth branch
    return (
        power_law(radius, x0[0], 1.0 / s4[0], c4[0]),
        x0,
        (1.0 / s4[0], s4[1] / s4[0]),
        c4,
    )


def otegi_2020_radius(
    mass: float,
    density: float = 0.0,
    bivariate: float = 0.5,
    silent: bool = False,
) -> Tuple[float, tuple, tuple, tuple]:
    r"""Calculate the radius of a planet using Otegi et al. (2020).

    Power law approximation:
        :math:`radius = x_0 C \times mass^S`
    Citation:
        Otegi, J. F., Bouchy, F., & Helled, R. 2020, A&A, 634, A43

    Parameters
    ----------
    mass : float
        Mass of the planet, in Earth radii.
    density : float, optional. Default: 0.0
        Density of the planet, in kg m^-3.
    bivariate : float, optional. Default: 0.5
        Probability that the returned mass that falls in the bivariate
        region is calculated with lower (rho >= 3300 kg m^-3) branch, instead
        of using the upper (rho < 3300 kg m^-3) branch of the power-law
        approximation. Must be a number between 0 and 1.
    silent : bool, optional. Default: False
        Whether to silence the warning if the radius is greater than the
        maximum value used by Otegi et al. (2020), or if the estimation falls
        in a multivariate region.

    Returns
    -------
    radius : float
        Radius of the planet, in Earth radii.
    c : tuple
        Constant of the power-law relation and its error.
    s : tuple
        Slope of the power-law relation and its error.
    x0 : tuple
        Reference value of the power-law and its error.
    """
    if mass > 120 and not silent:
        warnings.warn(
            "Radius is greater than the maximum value "
            + "used by Otegi et al. (2020): M = 120 M_earth.\n"
            + "The power-law approximation may not be accurate.",
            stacklevel=2,
        )

    # Otegi cuts at rho = 3300 kg m^-3 = 3.3 g cm^-3
    # Constants
    c1 = (1.03, 0.02)  # lower branch: >= 3300 kg m^-3
    c2 = (0.70, 0.11)  # upper branch: < 3300 kg m^-3
    # Slopes
    s1 = (0.29, 0.01)  # lower branch: >= 3300 kg m^-3
    s2 = (0.63, 0.04)  # upper branch: < 3300 kg m^-3
    # Reference value
    x0 = (1.0, 0.0)
    # Density cut
    dens_cut = 3300  # kg m^-3

    # If densityty is set
    if density > 0.0:  # If density is set
        if density >= dens_cut:  # Dense planet
            return power_law(mass, c1[0], s1[0], x0[0]), c1, s1, x0
        return power_law(mass, c2[0], s2[0], x0[0]), c2, s2, x0

    # Naive mass aproach
    if mass < 5:  # Small dense planet
        return power_law(mass, c1[0], s1[0], x0[0]), c1, s1, x0
    elif mass > 40:  # Large subdense planet
        return power_law(mass, c2[0], s2[0], x0[0]), c2, s2, x0

    # In density not set... Get dens_cut in [M_ear R_ear^-3]
    scaled_dens = convert(
        dens_cut, from_units=("kg", "m"), to_units=("me", "re"), power=(1, -3)
    )

    # Try both branches
    radius1 = power_law(mass, c1[0], s1[0], x0[0])
    radius2 = power_law(mass, c2[0], s2[0], x0[0])

    # Calculate the density
    density1 = mass / (4 / 3 * np.pi * radius1**3)
    density2 = mass / (4 / 3 * np.pi * radius2**3)

    # Check if lower or upper branch

    if density1 >= scaled_dens and density2 > scaled_dens:
        # First branch is valid
        return radius1, c1, s1, x0
    elif density1 < scaled_dens and density2 < scaled_dens:
        # Second branch is valid
        return radius2, c2, s2
    elif density1 >= scaled_dens and density2 < scaled_dens:
        if not silent:
            warnings.warn(
                "The estimation falls in a multivariate region. "
                + "The power-law approximation may not be accurate.",
                stacklevel=2,
            )
        # Both branches are valid.
        if not isinstance(bivariate, (int, float)) or (
            bivariate < 0 or bivariate > 1
        ):
            raise ValueError("Bivariate must be a number between 0 and 1.")

        if np.random.rand() < bivariate:  # Use first branch
            return radius1, c1, s1, x0
        return radius2, c2, s2, x0  # Use second branch

    # If nothing works, we use the closest branch
    if abs(density1 - scaled_dens) < abs(density2 - scaled_dens):
        # First branch is closer
        return radius1, c1, s1, x0

    # Second branch is closer
    return radius2, c2, s2, x0


def otegi_2020_mass(
    radius: float,
    density: float = 0.0,
    bivariate: float = 0.5,
    silent: bool = False,
) -> Tuple[float, tuple, tuple, tuple]:
    r"""Calculate the mass of a planet using Otegi et al. (2020).

    Power law approximation:
        :math:`mass = x_0 \frac{1}{C} \times radius^{1/S}`
    Citation:
        Otegi, J. F., Bouchy, F., & Helled, R. 2020, A&A, 634, A43

    Parameters
    ----------
    radius : float
        Mass of the planet, in Earth radii.
    density : float, optional. Default: 0.0
        Density of the planet, in kg m^-3.
    bivariate : float, optional. Default: 0.5
        Probability that the returned radius that falls in the bivariate
        region is calculated with lower (rho >= 3300 kg m^-3) branch, instead
        of using the upper (rho < 3300 kg m^-3) branch of the power-law
        approximation. Must be a number between 0 and 1.
    silent : bool, optional. Default: False
        Whether to print a warning if the radius is greater than the maximum
        value used by Otegi et al. (2020), or if the estimation falls
        in a multivariate region.

    Returns
    -------
    mass : float
        Mass of the planet, in Earth masses.
    inv_c : tuple
        Inverse of the constant of the power-law relation, and its error.
    inv_s : tuple
        Inverse of the slope of the power-law relation, and its error.
    x0 : tuple
        Reference value of the power-law, and its error.
    """
    if radius > 14.3 and not silent:
        warnings.warn(
            "Radius is greater than the maximum value "
            + "used by Otegi et al. (2020): R = 14.3 R_earth.\n"
            + "The power-law approximation may not be accurate.",
            stacklevel=2,
        )

    # Otegi cuts at rho = 3300 kg m^-3 = 3.3 g cm^-3
    # Constants
    c1 = (0.90, 0.06)  # lower branch: >= 3300 kg m^-3
    c2 = (1.74, 0.38)  # upper branch: < 3300 kg m^-3
    # Slopes
    s1 = (3.45, 0.12)  # lower branch: >= 3300 kg m^-3
    s2 = (1.58, 0.10)  # upper branch: < 3300 kg m^-3
    # Reference value
    x0 = (1.0, 0.0)
    # Density cut
    dens_cut = 3300  # kg m^-3

    # If densityty is set
    if density > 0.0:  # If density is set
        if density >= dens_cut:  # Dense planet
            return power_law(radius, c1[0], s1[0], x0[0]), c1, s1, x0
        return power_law(radius, c2[0], s2[0], x0[0]), c2, s2, x0

    # Naive radius aproach
    if radius < 1.5:  # Small dense planet
        return power_law(radius, c1[0], s1[0], x0[0]), c1, s1, x0
    elif radius > 3.1:  # Large subdense planet
        return power_law(radius, c2[0], s2[0], x0[0]), c2, s2, x0

    # In density not set... Get dens_cut in [M_ear R_ear^-3]
    scaled_dens = convert(
        dens_cut, from_units=("kg", "m"), to_units=("me", "re"), power=(1, -3)
    )

    # Try both branches
    mass1 = power_law(radius, c1[0], s1[0], x0[0])
    mass2 = power_law(radius, c2[0], s2[0], x0[0])

    # Calculate the density
    density1 = mass1 / (4 / 3 * np.pi * radius**3)
    density2 = mass2 / (4 / 3 * np.pi * radius**3)

    # Check if lower or upper branch

    if density1 >= scaled_dens and density2 > scaled_dens:
        # First branch is valid
        return mass1, c1, s1, x0
    elif density1 < scaled_dens and density2 < scaled_dens:
        # Second branch is valid
        return mass2, c2, s2, x0
    elif density1 >= scaled_dens and density2 < scaled_dens:
        if not silent:
            warnings.warn(
                "The estimation falls in a multivariate region. "
                + "The power-law approximation may not be accurate.",
                stacklevel=2,
            )
        # Both branches are valid.
        if not isinstance(bivariate, (int, float)) or (
            bivariate < 0 or bivariate > 1
        ):
            raise ValueError("Bivariate must be a number between 0 and 1.")

        if np.random.rand() < bivariate:  # Use first branch
            return mass1, c1, s1, x0
        return mass2, c2, s2, x0  # Use second branch

    # If nothing works, we use the closest branch
    if abs(density1 - scaled_dens) < abs(density2 - scaled_dens):
        # First branch is closer
        return mass1, c1, s1, x0

    # Second branch is closer
    return mass2, c2, s2, x0


def edmonson_2023_radius(mass: float) -> Tuple[float, tuple, tuple, tuple]:
    r"""Calculate the radius of a planet using the Edmondson et al. (2023).

    Power law approximation:
        :math:`radius = x_0 C \times mass^S`
    Citation:
        Edmondson, K., Norris, J., & Kerins, E. 2023, Open J. Astrophysics,
        submitted [arXiv:2310.16733]

    Parameters
    ----------
    mass : float
        Mass of the planet, in Earth masses.

    Returns
    -------
    radius : float
        Radius of the planet, in Earth radii.
    c : tuple
        Constant of the power-law relation and its error.
    s : tuple
        Slope of the power-law relation and its error.
    x0 : tuple
        Reference value of the power-law and its error.
    """
    # Constants
    c1 = (1.01, 0.03)
    c2 = (0.53, 0.05)
    c3 = (13, 1.2)
    # Slopes
    s1 = (0.28, 0.03)
    s2 = (0.68, 0.02)
    s3 = (0.012, 0.003)
    # Reference value
    x0 = (1.0, 0.0)
    # Transition mass
    m1_tr = 4.95
    m2_tr = 115

    if mass < m1_tr:  # First branch
        return power_law(mass, c1[0], s1[0], x0[0]), c1, s1, x0
    elif mass < m2_tr:  # Second branch
        return power_law(mass, c2[0], s2[0], x0[0]), c2, s2, x0
    # Third branch
    return power_law(mass, c3[0], s3[0], x0[0]), c3, s3, x0


def edmonson_2023_mass(radius: float) -> Tuple[float, tuple, tuple, tuple]:
    r"""Calculate the mass of a planet using the Edmondson et al. (2023).

    Power law approximation:
        :math:`mass = x_0 \frac{1}{C} \times radius^{1/S}`
    Citation:
        Edmondson, K., Norris, J., & Kerins, E. 2023, Open J. Astrophysics,
        submitted [arXiv:2310.16733]

    Parameters
    ----------
    radius : float
        Radius of the planet, in Earth radii.

    Returns
    -------
    mass : float
        Mass of the planet, in Earth masses.
    inv_c : tuple
        Inverse of the constant of the power-law relation, and its error.
    inv_s : tuple
        Inverse of the slope of the power-law relation, and its error.
    x0 : tuple
        Reference value of the power-law, and its error.
    """
    # Constants
    c1 = (1.01, 0.03)
    c2 = (0.53, 0.05)
    c3 = (13, 1.2)
    # Slopes
    s1 = (0.28, 0.03)
    s2 = (0.68, 0.02)
    s3 = (0.012, 0.003)
    # Reference value
    x0 = (1.0, 0.0)
    # Transition radius
    r1_tr = power_law(4.95, c1[0], s1[0], x0[0])  # 4.95 Earth masses
    r2_tr = power_law(115, c2[0], s2[0], x0[0])  # 115 Earth masses
    rmax = power_law(1e4, c3[0], s3[0], x0[0])  # 10000 earth masses

    # No multivariate region, because the power-law is always defined positive.

    if radius < r1_tr:  # First branch
        return (
            power_law(radius, x0[0], 1.0 / s1[0], c1[0]),
            x0,
            (1.0 / s1[0], s1[1] / s1[0]),
            c1,
        )
    elif radius < r2_tr:  # Second branch
        return (
            power_law(radius, x0[0], 1.0 / s2[0], c2[0]),
            x0,
            (1.0 / s2[0], s2[1] / s2[0]),
            c2,
        )
    elif radius > rmax:  # No estimation
        print(radius, rmax)
        raise ValueError(
            "Radius is greater than the maximum value used by "
            + f"Edmonson et al. (2023): R = {rmax} R_earth."
        )

    # Third branch
    return (
        power_law(radius, x0[0], 1.0 / s3[0], c3[0]),
        x0,
        (1.0 / s3[0], s3[1] / s3[0]),
        c3,
    )


def muller_2024_radius(mass: float) -> Tuple[float, tuple, tuple, tuple]:
    r"""Calculate the radius of a planet using the Müller et al. (2024).

    Power law approximation:
        :math:`radius = x_0 C \times mass^S`
    Citation:
        Müller S., Baron J., Helled R., Bouchy F. & Parc L. 2024, A&A, 686, A296

    Parameters
    ----------
    mass : float
        Mass of the planet, in Earth masses.

    Returns
    -------
    radius : float
        Radius of the planet, in Earth radii.
    c : tuple
        Constant of the power-law relation and its error.
    s : tuple
        Slope of the power-law relation and its error.
    x0 : tuple
        Reference value of the power-law and its error.
    """
    # Constants
    c1 = (1.02, 0.03)
    c2 = (0.56, 0.03)
    c3 = (18.6, 6.7)
    # Slopes
    s1 = (0.27, 0.04)
    s2 = (0.67, 0.05)
    s3 = (-0.06, 0.07)
    # Reference value
    x0 = (1.0, 0.0)
    # Transition mass
    m1_tr = 4.37
    m2_tr = 127

    if mass < m1_tr:  # First branch
        return power_law(mass, c1[0], s1[0], x0[0]), c1, s1, x0
    elif mass < m2_tr:  # Second branch
        return power_law(mass, c2[0], s2[0], x0[0]), c2, s2, x0
    # Third branch
    return power_law(mass, c3[0], s3[0], x0[0]), c3, s3, x0


def muller_2024_mass(
    radius: float, bivariate: float = 0.5, silent: bool = False
) -> Tuple[float, tuple, tuple, tuple]:
    r"""Calculate the mass of a planet using the Müller et al. (2024).

    Power law approximation:
        :math:`mass = x_0 \frac{1}{C} \times radius^{1/S}`
    Citation:
        Müller S., Baron J., Helled R., Bouchy F. & Parc L. 2024, A&A, 686, A296

    Parameters
    ----------
    radius : float
        Radius of the planet, in Earth radii.
    bivariate : float, optional. Default: 0.5
        Probability that the returned mass that falls in the bivariate
        region is calculated with the second (left) branch, instead of using
        the third (right) branch of the power-law approximation. Must be a
        number between 0 and 1.
    silent : bool, optional. Default: False
        Whether to silence the warning if the radius falls in the
        bivariate region.

    Returns
    -------
    mass : float
        Mass of the planet, in Earth masses.
    inv_c : tuple
        Inverse of the constant of the power-law relation, and its error.
    inv_s : tuple
        Inverse of the slope of the power-law relation, and its error.
    x0 : tuple
        Reference value of the power-law, and its error.
    """
    # Constants
    c1 = (1.02, 0.03)
    c2 = (0.56, 0.03)
    c3 = (18.6, 6.7)
    # Slopes
    s1 = (0.27, 0.04)
    s2 = (0.67, 0.05)
    s3 = (-0.06, 0.07)
    # Reference value
    x0 = (1.0, 0.0)
    # Transition radius
    r1_tr = 1.64
    r2_tr = power_law(127, c2[0], s2[0], x0[0])  # 127 Earth masses
    r3_tr = power_law(1e4, c3[0], s3[0], x0[0])  # Top: 1e4 Earth masses

    # We use the inverse of the constant and slope,
    # so the error is recaclulated with propagation error.

    if radius < r1_tr:  # First branch
        return (
            power_law(radius, x0[0], 1.0 / s1[0], c1[0]),
            x0,
            (1.0 / s1[0], s1[1] / s1[0]),
            c1,
        )
    elif radius < r3_tr:  # Pure second branch
        return (
            power_law(radius, x0[0], 1.0 / s2[0], c2[0]),
            x0,
            (1.0 / s2[0], s2[1] / s2[0]),
            c2,
        )
    elif radius > r2_tr:  # No estimation
        raise ValueError(
            "Radius is greater than the maximum value used by "
            + f"Müller et al. (2024): R = {r2_tr} R_earth."
        )

    # Bivariate region
    if not silent:
        warnings.warn(
            "Radius falls in the bivariate region: "
            + f"{r3_tr} < R < {r2_tr}"
            + "\n The mass-radius relation may not be accurate.",
            stacklevel=2,
        )
    if (not isinstance(bivariate, (int, float))) or (
        bivariate < 0 or bivariate > 1
    ):
        raise ValueError("Bivariate must be a number between 0 and 1.")

    if np.random.rand() < bivariate:  # Second branch
        return (
            power_law(radius, x0[0], 1.0 / s2[0], c2[0]),
            x0,
            (1.0 / s2[0], s2[1] / s2[0]),
            c2,
        )
    # Third branch
    return (
        power_law(radius, x0[0], 1.0 / s3[0], c3[0]),
        x0,
        (1.0 / s3[0], s3[1] / s3[0]),
        c3,
    )


def estimate_mass_single(
    radius: float,
    radius_err_min: float = 0.0,
    radius_err_max: float = 0.0,
    model: str = "ck17",
    multivariate: float = 0.5,
    err_method: int = 0,
    density: float = 0.0,
    silent: bool = False,
) -> Tuple[float, float, float]:
    r"""Calculate the mass of a planet using a power-law approximation.

    Equation:
        :math:`mass = \frac{1}{C} \times radius^{1/S}`

    Parameters
    ----------
    radius : float
        Radius of the planet, in Earth radii.
    radius_err_min : float
        Lower error of the radius, in Earth radii.
    radius_err_max : float
        Upper error of the radius, in Earth radii.
    model : str, optional. Default: "ck17"
        Model to use for the mass-radius power-law relation.
        'ck17': Chen & Kipping (2017) [trivariate]
        'o20': Otegi et al. (2020) [density|bivariate]
        'e23': Edmondson et al. (2023)
        'm24': Müller et al. (2024) [bivariate]
    multivariate : float, tuple, optional. Default: 0.5
        Probability of using the (first, second, ...) branch if the estimation
        falls in a multivariate region.
        For bivariate models ('o20', 'm24'), it must be a float between
        0 and 1.
        For trivariate model "ck17", it must be a tuple of two floats between
        0 and 1, where the sum of them must be lower equal than 1.
    err_method : int, optional. Default: 0
        Which method implement for error calculation.
        Method 0: Do not calculate errors. Return both as 0.0.
        Method 1: (Naive) Error propagation with the power-law approximation,
        using the radius error as the maximum of the two extremes.
        Warning: May return excessively large errors for multivariate
        sections.
        Method 2: Evaluate the radius extremes and calculate each mass
        extreme with the power-law approximation.
        Method 3: Returns the approximate model error as value errors.
    density : float, optional. Default: 0.0
        Density of the planet, in kg m^-3.
        Only used if model is 'o20'. If equal to 0.0, the code uses
        multivariate float instead, to determine which branch to use.
    silent : bool, optional. Default: False
        Whether to silence the warning if the radius falls in a
        multivariate region, or if the estimation is not accurate.

    Returns
    -------
    mass : float
        Mass of the planet, in Earth masses.
    mass_err_min : float
        Lower error of the mass, in Earth masses. If err_method=0, it is 0.0.
    mass_err_max : float
        Upper error of the mass, in Earth masses. If err_method=0, it is 0.0.
    """
    # Check if radius is NaN
    if np.isnan(radius):
        return np.nan, np.nan, np.nan
    # Calculate the mass
    if model == "ck17":
        mass, c, s, x0 = chen_kipp_2017_mass(
            radius, trivariate=multivariate, silent=silent
        )
    elif model == "o20":
        mass, c, s, x0 = otegi_2020_mass(
            radius, density=density, bivariate=multivariate, silent=silent
        )
    elif model == "e23":
        mass, c, s, x0 = edmonson_2023_mass(radius)
    elif model == "m24":
        mass, c, s, x0 = muller_2024_mass(
            radius, bivariate=multivariate, silent=silent
        )
    else:
        raise ValueError(
            f"Model {model} not implemented. Use 'ck17', 'o20', 'e23' or 'm24'."
        )

    # Calculate the mass error
    if err_method == 0:
        return mass, 0.0, 0.0

    # Warn if necessary
    if not silent and model in ["ck17", "o20", "m24"] and err_method == 1:
        warnings.warn(
            "Using the naive error propagation method may generate"
            + " excessively large errors in multivariate sections",
            stacklevel=2,
        )

    # Use auxiliar error function
    mass_err_min, mass_err_max = _aux_error_estimator(
        radius,
        radius_err_max,
        radius_err_min,
        mass,
        c,
        s,
        x0,
        err_method,
        silent,
        0,
    )

    return mass, mass_err_min, mass_err_max


def estimate_radius_single(
    mass: float,
    mass_err_min: float = 0.0,
    mass_err_max: float = 0.0,
    model: str = "ck17",
    bivariate: float = 0.5,
    err_method: int = 0,
    density: float = 0.0,
    silent: bool = False,
) -> Tuple[float, float, float]:
    r"""Calculate the radius of a planet using the power-law approximation.

    Equation:
        :math:`radius = C \times mass^S`

    Parameters
    ----------
    mass : float
        Mass of the planet, in Earth masses.
    mass_err_min : float
        Lower error of the mass, in Earth masses.
    mass_err_max : float
        Upper error of the mass, in Earth masses.
    model : str, optional. Default: "ck17"
        Model to use for the mass-radius power-law relation.
        'ck17': Chen & Kipping (2017)
        'o20': Otegi et al. (2020) [density|bivariate]
        'e23': Edmondson et al. (2023)
        'm24': Müller et al. (2024)
    bivariate : float, optional. Default: 0.5
        Probability of using the lower branch if the estimation falls in a
        bivariate region. Must be a number between 0 and 1.
        Only used if model is 'o20'.
    err_method : int, optional. Default: 0
        Which method implement for error calculation.
        Method 0: Do not calculate errors. Return both as 0.0.
        Method 1: (Naive) Error propagation with the power-law approximation,
        using the mass error as the maximum of the two extremes.
        Method 2: Evalaute the mass extremes and calculate each radius extreme.
        Method 3: Returns the approximate model error as value errors.
    density : float, optional. Default: 0.0
        Density of the planet, in kg m^-3.
        Only used if model is 'o20'. If equal to 0.0, the code uses
        bivariate float value instead, to determine which branch to use.
    silent : bool, optional. Default: False
        Whether to silence the warning if the radius falls in a
        bivariate region, or if the estimation is not accurate.

    Returns
    -------
    radius : float
        Radius of the planet, in Earth radii.
    radius_err_min : float
        Lower error of the radius, in Earth radii. If err_method=0, it is 0.0.
    radius_err_max : float
        Upper error of the radius, in Earth radii. If err_method=0, it is 0.0.
    """
    # Check if mass is NaN
    if np.isnan(mass):
        return np.nan, np.nan, np.nan
    # Calculate the radius
    if model == "ck17":
        radius, c, s, x0 = chen_kipp_2017_radius(mass)
    elif model == "o20":
        radius, c, s, x0 = otegi_2020_radius(
            mass, density=density, bivariate=bivariate, silent=silent
        )
    elif model == "e23":
        radius, c, s, x0 = edmonson_2023_radius(mass)
    elif model == "m24":
        radius, c, s, x0 = muller_2024_radius(mass)
    else:
        raise ValueError(
            f"Model {model} not implemented. Use 'ck17', 'o20', 'e23' or 'm24'."
        )

    # Calculate the radius error
    if err_method == 0:
        return radius, 0.0, 0.0

    # Warn if necessary
    if not silent and model == "o20" and err_method == 1:
        warnings.warn(
            "Using the naive error propagation method may generate"
            + " excessively large errors in multivariate sections",
            stacklevel=2,
        )

    # Use auxiliar error function
    radius_err_min, radius_err_max = _aux_error_estimator(
        mass,
        mass_err_max,
        mass_err_min,
        radius,
        c,
        s,
        x0,
        err_method,
        silent,
        1,
    )

    return radius, radius_err_min, radius_err_max


def _aux_error_estimator(
    val: float,
    val_err_max: float,
    val_err_min: float,
    output: float,
    c: tuple,
    s: tuple,
    x0: tuple,
    method: int,
    silent: bool,
    which: int,
) -> Tuple[float, float]:
    """Auxiliary function to estimate the error of a power-law relation."""
    # Handle errors as absolute values
    val_err_min = abs(val_err_min)
    val_err_max = abs(val_err_max)

    # Calculate the output error
    if method == 1:  # Naive error propagation
        val_err = max(val_err_min, val_err_max)
        output_err_max = power_law_error(
            val, val_err, c[0], c[1], s[0], s[1], x0[0], x0[1], y=output
        )
        output_err_min = -output_err_max
    elif method == 2:  # Calculate the error at output extremes
        if not silent and any([val_err_min, val_err_max]) == 0.0:
            txt = ["mass", "radius"] if which == 1 else ["radius", "mass"]
            warnings.warn(
                f"Calculating the {txt[0]} error at extremes without "
                + f"a {txt[1]} error generates no {txt[0]} error",
                stacklevel=2,
            )
        output_min = power_law(val - val_err_min, c[0], s[0], x0[0])
        output_max = power_law(val + val_err_max, c[0], s[0], x0[0])
        # Calculate the output error. Safe sign
        output_err_min = min(min(output_min, output_max), output) - output
        output_err_max = max(max(output_min, output_max), output) - output
    elif (
        method == 3
    ):  # Give the error from the model [eg.: y+ = pw(x, c+, s+, x0-]
        output_min = power_law(val, c[0] - c[1], s[0] - s[1], x0[0] + x0[1])
        output_max = power_law(val, c[0] + c[1], s[0] + s[1], x0[0] - x0[1])
        # Calculate the output error. Safe sign
        output_err_min = min(min(output_min, output_max), output) - output
        output_err_max = max(max(output_min, output_max), output) - output
    else:
        raise ValueError(f"Error method '{method}' not implemented.")

    return output_err_min, output_err_max


estimate_mass_vec = np.vectorize(
    estimate_mass_single,
    doc="Vectorized version of :py:func:`estimate_mass_single`.",
    excluded=["model", "multivariate", "err_method", "density", "silent"],
)


estimate_radius_vec = np.vectorize(
    estimate_radius_single,
    doc="Vectorized version of :py:func:`estimate_radius_single`.",
    excluded=["model", "bivariate", "err_method", "density", "silent"],
)


def estimate_radius(
    mass: Union[float, np.ndarray],
    mass_err_min: Union[float, np.ndarray] = 0.0,
    mass_err_max: Union[float, np.ndarray] = 0.0,
    model: str = "ck17",
    bivariate: float = 0.5,
    err_method: int = -1,
    density: float = 0.0,
    silent: bool = False,
) -> Union[Tuple[float, float, float], np.ndarray]:
    r"""Calculate the radius of a planet using the power-law approximation.

    The different models available are:
    Chen & Kipping (2017),
    Otegi et al. (2020),
    Edmondson et al. (2023),
    and Müller et al. (2024).

    Equation:
        :math:`radius = C \times mass^S`

    Parameters
    ----------
    mass : float, np.ndarray
        Mass of the planet, in Earth masses.
    mass_err_min : float, np.ndarray
        Lower error of the mass, in Earth masses.
    mass_err_max : float, np.ndarray
        Upper error of the mass, in Earth masses.
    model : str, optional. Default: "ck17"
        Model to use for the mass-radius power-law relation.
        'ck17': Chen & Kipping (2017)
        'o20': Otegi et al. (2020) [density|bivariate]
        'e23': Edmondson et al. (2023)
        'm24': Müller et al. (2024)
    bivariate : float, optional. Default: 0.5
        Probability of using the lower branch if the estimation falls in a
        bivariate region. Must be a number between 0 and 1.
        Only used if model is 'o20'.
    err_method : int, optional. Default: -1
        Which method implement for error calculation.
        Method -1: Do not calculate errors. Return only the radius.
        Method 0: Do not calculate errors. Return both as 0.0.
        Method 1: (Naive) Error propagation with the power-law approximation,
        using the mass error as the maximum of the two extremes.
        Method 2: Evalaute the mass extremes and calculate each radius extreme.
        Method 3: Returns the approximate model error as value errors.
    density : float, optional. Default: 0.0
        Density of the planet, in kg m^-3.
        Only used if model is 'o20'. If equal to 0.0, the code uses
        bivariate
    silent : bool, optional. Default: False
        Whether to silence the warning if the radius falls in a
        bivariate region, or if the estimation is not accurate.

    Returns
    -------
    result : tuple[float, float, float] or np.ndarray
        Estimated radius, and its minimum and maximum errors, in Earth radii.
        If mass is a scalar, the tuple is
        (radius, radius_err_min, radius_err_max),
        else it is a (n,3) numpy array.
        If `err_method=0`, the tuple | array is (radius, 0.0, 0.0).

    References
    ----------
        `Chen, J., & Kipping, D. 2017, ApJ, 834, 17
        <https://ui.adsabs.harvard.edu/abs/2017ApJ...834...17C>`_

        `Otegi, J. F., Bouchy, F., & Helled, R. 2020, A&A, 634, A43
        <https://ui.adsabs.harvard.edu/abs/2020A&A...634A..43O>`_

        `Edmondson, K., Norris, J., & Kerins, E. 2023, Open J. Astrophysics,
        submitted <https://arxiv.org/abs/2310.16733>`_

        `Müller S., Baron J., Helled R., Bouchy F. & Parc L. 2024, A&A, 686,
        A296 <https://ui.adsabs.harvard.edu/abs/2024A&A...686A.296M>`_
    """
    err_method_aux = 0 if err_method == -1 else err_method
    if isinstance(mass, (int, float)):
        radius, radius_err_min, radius_err_max = estimate_radius_single(
            mass=mass,
            mass_err_min=mass_err_min,
            mass_err_max=mass_err_max,
            model=model,
            bivariate=bivariate,
            err_method=err_method_aux,
            density=density,
            silent=silent,
        )
    else:
        radius, radius_err_min, radius_err_max = estimate_radius_vec(
            mass=mass,
            mass_err_min=mass_err_min,
            mass_err_max=mass_err_max,
            model=model,
            bivariate=bivariate,
            err_method=err_method_aux,
            density=density,
            silent=silent,
        )

    # Return only the radius?
    if err_method == -1:
        return radius

    return np.array(
        [radius, -radius_err_min, radius_err_max]
    ).T  # -for consistency


def estimate_mass(
    radius: Union[float, np.ndarray],
    radius_err_min: Union[float, np.ndarray] = 0.0,
    radius_err_max: Union[float, np.ndarray] = 0.0,
    model: str = "ck17",
    multivariate: Union[float, tuple, list] = None,
    err_method: int = -1,
    density: float = 0.0,
    silent: bool = False,
) -> Union[Tuple[float, float, float], np.ndarray]:
    r"""Calculate the mass of a planet using a power-law approximation.

    The different models available are:
    Chen & Kipping (2017),
    Otegi et al. (2020),
    Edmondson et al. (2023),
    and Müller et al. (2024).

    Equation:
        :math:`mass = \frac{1}{C} \times radius^{1/S}`

    Parameters
    ----------
    radius : float, np.ndarray
        Radius of the planet, in Earth radii.
    radius_err_min : float, np.ndarray
        Lower error of the radius, in Earth radii.
    radius_err_max : float, np.ndarray
        Upper error of the radius, in Earth radii.
    model : str, optional. Default: "ck17"
        Model to use for the mass-radius power-law relation.
        'ck17': Chen & Kipping (2017) [trivariate]
        'o20': Otegi et al. (2020) [density|bivariate]
        'e23': Edmondson et al. (2023)
        'm24': Müller et al. (2024) [bivariate]
    multivariate : float, tuple, optional. Default: 0.5
        Probability of using the (first, second, ...) branch if the estimation
        falls in a multivariate region.
        For bivariate models ('o20', 'm24'), it must be a float between
        0 and 1. Default is 0.5.
        For trivariate model "ck17", it must be a tuple of two floats between
        0 and 1, where the sum of them must be lower equal than 1. Default is
        (0.1, 0.85).
    err_method : int, optional. Default: -1
        Which method implement for error calculation.
        Method -1: Do not calculate errors. Return only the mass.
        Method 0: Do not calculate errors. Return both as 0.0.
        Method 1: (Naive) Error propagation with the power-law approximation,
        using the radius error as the maximum of the two extremes.
        Warning: May return excessively large errors for multivariate
        sections.
        Method 2: Evaluate the radius extremes and calculate each mass
        extreme with the power-law approximation.
        Method 3: Returns the approximate model error as value errors.
    density : float, optional. Default: 0.0
        Density of the planet, in kg m^-3.
        Only used if model is 'o20'. If equal to 0.0, the code uses
        multivariate float instead, to determine which branch to use.
    silent : bool, optional. Default: False
        Whether to silence the warning if the radius falls in a
        multivariate region, or if the estimation is not accurate.

    Returns
    -------
    result : tuple[float, float, float] or np.ndarray
        Estimated mass, and its minimum and maximum errors, in Earth masses.
        If radius is a scalar, the tuple is (mass, mass_err_min, mass_err_max),
        else it is a (n,3) numpy array.
        If `err_method=0`, the tuple | array is (mass, 0.0, 0.0).

    References
    ----------
        `Chen, J., & Kipping, D. 2017, ApJ, 834, 17
        <https://ui.adsabs.harvard.edu/abs/2017ApJ...834...17C>`_

        `Otegi, J. F., Bouchy, F., & Helled, R. 2020, A&A, 634, A43
        <https://ui.adsabs.harvard.edu/abs/2020A&A...634A..43O>`_

        `Edmondson, K., Norris, J., & Kerins, E. 2023, Open J. Astrophysics,
        submitted <https://arxiv.org/abs/2310.16733>`_

        `Müller S., Baron J., Helled R., Bouchy F. & Parc L. 2024, A&A, 686,
        A296 <https://ui.adsabs.harvard.edu/abs/2024A&A...686A.296M>`_
    """
    err_method_aux = 0 if err_method == -1 else err_method
    if multivariate is None:  # Set default values
        if model == "ck17":
            multivariate = (0.1, 0.85)
        else:
            multivariate = 0.5
    if isinstance(radius, (int, float)):
        mass, mass_err_min, mass_err_max = estimate_mass_single(
            radius=radius,
            radius_err_min=radius_err_min,
            radius_err_max=radius_err_max,
            model=model,
            multivariate=multivariate,
            err_method=err_method_aux,
            density=density,
            silent=silent,
        )
    else:
        mass, mass_err_min, mass_err_max = estimate_mass_vec(
            radius=radius,
            radius_err_min=radius_err_min,
            radius_err_max=radius_err_max,
            model=model,
            multivariate=multivariate,
            err_method=err_method_aux,
            density=density,
            silent=silent,
        )

    # Return only the mass?
    if err_method == -1:
        return mass

    return np.array([mass, -mass_err_min, mass_err_max]).T  # -for consistency



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# =============================================================================
# DOCS
# =============================================================================

"""Module with utilities for unit conversion and manipulation."""

# =============================================================================
# IMPORTS
# =============================================================================

from types import MappingProxyType
from typing import List, Tuple, Union

from numpy import pi

# =============================================================================
# CONSTANTS AND NORMALIZED UNIT DICTIONARIES
# =============================================================================

# Gravitational constant in SI units
G = 6.67430e-11  # m^3 kg^-1 s^-2

# UNITS dictionary
_units = {
    "mass": {
        "g": 1e-3,  # alias
        "gr": 1e-3,  # grams → kg
        "kg": 1.0,
        "ton": 1e3,  # metric ton → kg
        "me": 5.9722e24,  # Earth mass → kg
        "mj": 1.89813e27,  # Jupiter mass → kg
        "ms": 1.989e30,  # Solar mass → kg
    },
    "distance": {
        "cm": 1e-2,  # cm → m
        "m": 1.0,
        "km": 1e3,  # km → m
        "re": 6.371e6,  # Earth eq radius → m
        "rj": 6.9911e7,  # Jupiter radius → m
        "rs": 6.957e8,  # Solar radius → m
        "au": 1.495978e11,  # Astronomical unit → m
        "pc": 3.0857e16,  # Parsec → m
    },
    "time": {
        "sec": 1.0,
        "s": 1.0,  # alias
        "min": 60.0,  # min → s
        "hour": 3600.0,  # hour → s
        "day": 86400.0,  # day → s
        "year": 31557600.0,  # year → s
        "yr": 31557600,  # alias 2
    },
    "angle": {
        "rad": 1.0,
        "deg": pi / 180.0,
    },
}
# Add density units
_units["density"] = {
    "rhow": _units["mass"]["g"] / _units["distance"]["cm"] ** 3,
    "rhos": _units["mass"]["ms"]
    / (4.0 / 3.0 * pi * _units["distance"]["rs"] ** 3),
    "rhoj": _units["mass"]["mj"]
    / (4.0 / 3.0 * pi * _units["distance"]["rj"] ** 3),
    "rhoe": _units["mass"]["me"]
    / (4.0 / 3.0 * pi * _units["distance"]["re"] ** 3),
}

# Create immutable dict
UNITS = MappingProxyType(_units)

# Custom with MKS
_mks = {u: v for uv in UNITS.values() for u, v in uv.items()}
_mks["G"] = G

# Create dict to set MKS
MKS = MappingProxyType(_mks)


def _convert_units(from_unit, to_unit, power=1):
    from_unit = from_unit.lower()
    to_unit = to_unit.lower()

    for unit_dict in UNITS.values():
        if from_unit in unit_dict and to_unit in unit_dict:
            origin = unit_dict[from_unit]
            dest = unit_dict[to_unit]
            return (origin / dest) ** power

    raise ValueError(
        f"Cannot convert '{from_unit}' to '{to_unit}': incompatible or unknown."
    )


def __normalize_units(u):
    if isinstance(u, str):
        return [u]
    return u


def __normalize_powers(p, n):
    if isinstance(p, int):
        return [p] * n
    if len(p) != n:
        raise ValueError(
            f"Power length {len(p)} does not match units length {n}."
        )
    return p


def convert(
    *values,
    from_units: Union[None, str, Tuple[str, ...], List[str]] = None,
    to_units: Union[None, str, Tuple[str, ...], List[str]] = None,
    power: Union[int, Tuple[int, ...], List[int]] = 1,
) -> Union[float, List[float]]:
    """
    Convert between compound units (e.g. km^2/s → m^2/s).

    Parameters
    ----------
    from_units : str or list of str
        Units to convert from.
    to_units : str or list of str
        Units to convert to.
    power : int or list of int, optional
        Power(s) for each unit (default = 1).
        Must match length if list.

    Examples
    --------
    To convert 5 km s^{-2} to au yr^{-2}:

    >>> resokit.units.convert(
    ...     5,
    ...     from_units=("km", "s"),
    ...     to_units=("au", "yr"),
    ...     power=(1, -2)
    ... )
    33247713.903743323

    Returns
    -------
    float or list of float
        Converted value, or list of converted values.
    """
    if from_units is None:
        raise ValueError("Argument 'from_units' must be set.")
    if to_units is None:
        raise ValueError("Argument 'to_units' must be set.")

    from_units = __normalize_units(from_units)
    to_units = __normalize_units(to_units)
    powers = __normalize_powers(power, len(from_units))

    if len(from_units) != len(to_units):
        raise ValueError("Mismatch between number of from_units and to_units.")

    factor = 1.0
    for fu, tu, pw in zip(from_units, to_units, powers):
        factor *= _convert_units(fu, tu, power=pw)

    if not values:
        return factor
    if len(values) == 1:
        return factor * values[0]
    return [value * factor for value in values]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# This file indicates that the directory should be treated as a package.

# ============================================================================
# DOCS
# ============================================================================

"""The ResoKit.utils.mmr package includes diverse tools for MMRs analysis."""

# =============================================================================
# IMPORTS
# =============================================================================

from .mmrs import closest_mmr3b  # noqa
from .mmrs import label_mmr3b  # noqa
from .mmrs import mindist_mmr3b  # noqa
from .mmrs import mmr3b  # noqa
from .mmrs import mmrs_in_area  # noqa
from .mmrs import plot_mmrs  # noqa

# Make the functions available at the package level

__all__ = [
    "closest_mmr3b",
    "label_mmr3b",
    "mindist_mmr3b",
    "mmr3b",
    "mmrs_in_area",
    "plot_mmrs",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Module to manage 3-body mean-motion resonances (MMRs).

This module provides tools for identifying, labeling, and computing distances
to MMRs in the phase space.
"""

# ============================================================================
# IMPORTS
# ============================================================================

import warnings
from itertools import product
from typing import List, Tuple, Union

import matplotlib.pyplot as plt

import numpy as np

from scipy.optimize import minimize

# ============================================================================
# FUNCTIONS
# ============================================================================


def mmr3b(
    x: Union[float, np.ndarray], resonance: Tuple[int, int, int]
) -> Union[float, np.ndarray]:
    """Compute the 3-body mean-motion resonance (MMR) curve.

    Equation
        :math:`a * x + b + c / y = 0`

    Parameters
    ----------
    x : float or np.ndarray
        The independent variable for the curve.
    resonance : tuple[int, int, int]
        Coefficients (a, b, c) defining the resonance.

    Returns
    -------
    y : float or np.ndarray
        The corresponding values of the 3-body resonance curve.
        Singularities are replaced with NaN.
    """
    a, b, c = resonance  # Coefficients of the resonance curve

    if np.ndim(x) == 0:  # Single value
        if a * x + b == 0:
            return np.nan
        return -c / (a * x + b)

    # Avoid division by zero
    curve = np.divide(
        -c, (a * x + b), out=np.full_like(x, np.nan), where=(a * x + b) != 0
    )

    # Handle singularities
    singularity = -b / a
    if (
        np.ndim(x) > 0  # x is an array
        and singularity >= np.min(x)  # Singularity is within bounds
        and singularity <= np.max(x)  # Singularity is within bounds
    ):
        closest_idx = np.argmin(np.abs(x - singularity))
        curve[closest_idx] = np.nan

    return curve


def mmrs_in_area(
    bounds: Tuple[float, float, float, float],
    order3: int = 0,
    max_coeff3: int = 10,
    max_order3: int = 0,
    mmr2b: bool = True,
    order2: int = 2,
    max_coeff2: int = 10,
    max_order2: int = 2,
    verbose: bool = False,
) -> list:
    """Identify 3-body and 2-body MMRs in a specified phase-space region.

    This function identifies 3-body mean-motion resonances (3P-MMRs) and
    optionally 2-body mean-motion resonances (2P-MMRs) in a specified
    region of the phase space.

    Parameters
    ----------
    bounds : tuple[float, float, float, float]
        The limits of the region (x_min, x_max, y_min, y_max) in the phase
        space. Remeber that x_min and y_min must be > 1.
    order3 : int, optional. Default: 0
        Exact order for 3P-MMRs.
    max_coeff3 : int, optional. Default: 10
        Calculate 3P-MMRs up to this maximum integer coefficient.
    max_order3 : int, optional. Default: 0
        Calculate 3P-MMRs up to this maximum order.
        If set to 0, only the exact order is considered.
    mmr2b : bool, optional. Default: True
        Whether to compute 2P-MMRs.
    order2 : int, optional. Default: 2
        Exact order for 2P-MMRs.
    max_coeff2 : int, optional. Default: 10
        Calculate 2P-MMRs up to this maximum integer coefficient.
    max_order2 : int, optional. Default: 2
        Calculate 2P-MMRs up to this maximum order.
        If set to 0, only the exact order is considered.
    verbose : bool, optional. Default: False
        Whether to print the resonances found.

    Returns
    -------
    resonances: list
        A list containing detected 3P-MMRs and optionally
        2P-MMRs along the x and y axes. This last case is
        returned as [r3, r2x, r2y], where r3, r2x, and r2y
        are lists of 3P-MMRs, 2P-MMRs along the x-axis, and
        2P-MMRs along the y-axis, respectively.
    """
    # Get the bounds
    x_min, x_max, y_min, y_max = bounds

    # Check bounds
    if x_min < 1 or y_min < 1:
        raise ValueError("Bounds must be greater or equal than 1.")
    if x_min >= x_max or y_min >= y_max:
        raise ValueError("Invalid bounds. Must be ascending.")

    # Initialize the list of resonances
    r3p_resonances = []

    # Define the range of coefficients
    coeff_range = np.flip(np.arange(-max_coeff3, max_coeff3 + 1))

    # Define good_order function
    def good_order3(i, j, k):
        if max_order3 == 0:
            return abs(i + j + k) == order3
        return (i + j + k) <= max_order3

    # Identify 3P-MMRs
    for i in range(1, max_coeff3 + 1):
        for j, k in product(coeff_range, repeat=2):

            if (
                not good_order3(i, j, k)  # Check order
                # or i == 0  # Adjacent 2P-MMR
                or k == 0  # Adjacent 2P-MMR
                or (
                    j == 0 and not (i > 0 and k < 0)
                )  # Take (i,0,-k) over (-i,0,k)
            ):
                continue

            # Normalize coefficients
            gcd = np.gcd.reduce([i, j, k])
            i_r, j_r, k_r = i // gcd, j // gcd, k // gcd

            # Skip if the resonance is already in the list
            if [i_r, j_r, k_r] in r3p_resonances:
                continue

            # Check bounds for the resonance curve
            if not _is_curve_within_bounds([i_r, j_r, k_r], bounds):
                continue

            r3p_resonances.append([i_r, j_r, k_r])

    if verbose:
        print(f"Found {len(r3p_resonances)} 3-body mean-motion resonances.")

    # Check if 2P-MMRs are not required
    if not mmr2b:  # Return 3P-MMRs only
        return r3p_resonances

    # 2P-MMRs ideification  required

    # Define good_order function
    def good_order2(i, j):
        if max_order2 == 0:
            return (i - j) == order2
        return (i - j) <= max_order2

    r2p_x, r2p_y = [], []
    for i in range(2, max_coeff2 + 1):
        for j in range(1, i):

            # Check order
            if not good_order2(i, j):
                continue

            # Normalize coefficients
            gcd = np.gcd(i, j)
            i_r, j_r = i // gcd, j // gcd

            if x_min <= i_r / j_r <= x_max and [i_r, j_r] not in r2p_x:
                r2p_x.append([i_r, j_r])

            if y_min <= i_r / j_r <= y_max and [i_r, j_r] not in r2p_y:
                r2p_y.append([i_r, j_r])

    if verbose:
        print(
            f"Found {len(r2p_x)} 2-body mean-motion resonances "
            + "along the x-axis."
        )
        print(
            f"Found {len(r2p_y)} 2-body mean-motion resonances "
            + "along the y-axis."
        )

    return [r3p_resonances, r2p_x, r2p_y]


def _is_curve_within_bounds(
    resonance: List[int], bounds: Tuple[float, float, float, float]
) -> bool:
    """Determine if a resonance curve intersects a bounded region.

    Parameters
    ----------
    resonance : list[int, int, int]
        Coefficients defining the resonance.
    bounds : tuple[float, float, float, float]
        The bounding region as (x_min, x_max, y_min, y_max).

    Returns
    -------
    bool
        `True` if the curve intersects the region, `False` otherwise.
    """
    x_min, x_max, y_min, y_max = bounds
    i, j, k = resonance

    # No singularity handling needed
    if (-j / i < x_min) or (-j / i > x_max):

        # Si cruza el eje izquierdo
        if (-k / (i * x_min + j) >= y_min) and (-k / (i * x_min + j) <= y_max):
            return True

        # Si cruza el eje derecho
        elif (-k / (i * x_max + j) >= y_min) and (
            -k / (i * x_max + j) <= y_max
        ):
            return True

        # Si cruza el eje de abajo
        elif (-(j * y_min + k) / i / y_min >= x_min) and (
            -(j * y_min + k) / i / y_min <= x_max
        ):
            return True

        return False

    # Handle singularities
    else:

        # Si cruza el eje izquierdo
        if (
            not np.isclose(-j / i, x_min)
            and (-k / (i * x_min + j) >= y_min)
            and (-k / (i * x_min + j) <= y_max)
        ):
            return True

        # si cruza el eje derecho
        elif (
            not np.isclose(-j / i, x_max)
            and (-k / (i * x_max + j) >= y_min)
            and (-k / (i * x_max + j) <= y_max)
        ):
            return True

        # si cruza el eje de abajo
        elif (
            (-(j * y_min + k) / i / y_min >= x_min)
            and (-(j * y_min + k) / i / y_min < -j / i)
        ) or (
            (-(j * y_min + k) / i / y_min > -j / i)
            and (-(j * y_min + k) / i / y_min <= x_max)
        ):
            return True

        return False


def mindist_mmr3b(
    a: float,
    b: float,
    resonance: Tuple[int, int, int],
    x0: Union[float, None] = None,
    unphysical: bool = False,
    **minimize_kwargs,
) -> Tuple[float, float, float]:
    """Calculate the minimum distance to a 3-body resonance curve.

    Parameters
    ----------
    a : float
        The x-coordinate of the point.
    b : float
        The y-coordinate of the point.
    resonance : Tuple[int, int, int]
        Coefficients defining the resonance.
    x0 : float, optional. Default: None
        Initial guess for the optimization.
        If None, the function will use the middle point of the curve.
    unphysical : bool, optional. Default: False
        Whether to allow unphysical solutions. (y below 1)
    minimize_kwargs : dict, optional
        Additional arguments for :py:func:scipy.optimize.minimize
        function.

    Returns
    -------
    x_min : float
        The x coordinate of the minimum distance.
    y_min : float
        The y coordinate of the minimum distance.
    distance_min : float
        The minimum distance to the resonance curve.
    """
    # Singularity handling
    singularity = -resonance[1] / resonance[0]

    # Use the right or left side of the curve, in relation to singularity
    use_right = resonance[2] < 0
    # if "use_right", then the curve behaves like "1/x", else like "- 1/x"
    x_y1 = -(resonance[1] + resonance[2]) / resonance[0]

    # Fast check for (1,1) solutions
    if use_right and mmr3b(1, resonance) == 1 and not unphysical:
        # Here, the only physical solution is (1,1)
        return 1, 1, np.sqrt((1 - a) ** 2 + (1 - b) ** 2)

    # Avoid crossing the singularity
    if x_y1 < max(1, singularity):
        x_y1 = np.inf

    # Define the bounds for the optimization
    if use_right:
        bounds_x = [singularity, x_y1]
    elif not unphysical:
        bounds_x = [1, singularity]
    else:
        bounds_x = [-np.inf, singularity]

    # Function to calculate the r3p curve value at x
    def dist2_to_curve(x, unphysical=unphysical):
        y = mmr3b(x, resonance)  # Calculate the curve value
        if np.isnan(y) or (y < 1 and not unphysical):
            return np.inf
        return (x - a) ** 2 + (y - b) ** 2

    # Redefine x0 if necessary
    if x0 is None:
        # Use the middle point if the singularity is infinite
        if use_right:
            # Use the middle point if finite, else use the singularity + 1e-3
            x0 = (
                0.5 * (singularity + x_y1)
                if np.isfinite(x_y1)
                else singularity + 1e-3
            )
        else:
            x0 = 0.5 * (1 + singularity)  # Use the middle point

    elif x0 < bounds_x[0] or x0 > bounds_x[1]:  # Check if x0 is within bounds
        raise ValueError(f"Initial guess {x0} is out of bounds: {bounds_x}")

    # Function to calculate the distance between a point and the curve
    # Avoid runtime warnings in subtraction
    with np.errstate(invalid="ignore"):
        result = minimize(dist2_to_curve, x0, **minimize_kwargs)

    if result.success:
        x_min = result.x[0]
        distance_min = np.sqrt(result.fun)
        return x_min, mmr3b(x_min, resonance), distance_min
    else:
        raise ValueError("Optimization failed!")


def closest_mmr3b(
    a: float,
    b: float,
    order3: int = 0,
    max_coeff3: int = 10,
    max_order3: int = 0,
    bounds: Tuple[float, float, float, float] = None,
    radius: float = 1e-3,
    verbose: bool = True,
    **minimize_kwargs,
) -> Tuple[List[int], float]:
    """Find the closest 3-body mean-motion resonance to a point.

    Parameters
    ----------
    a : float
        The x-coordinate of the point.
    b : float
        The y-coordinate of the point.
    order3 : int, optional. Default: 0
        Exact order for 3P-MMRs.
    max_coeff3 : int, optional. Default: 10
        Calculate 3P-MMRs up to this maximum integer coefficient.
    max_order3 : int, optional. Default: 0
        Calculate 3P-MMRs up to this maximum order.
        If set to 0, only the exact order is considered.
    bounds : tuple[float, float, float, float], optional. Default: None
        The limits of the region (x_min, x_max, y_min, y_max).
        If not provided and radius is not None, the function will
        use the search radius around the point. If
    radius : float, optional. Default: 1e-3
        The radius of the search area around the point.
        If bounds are provided, the radius will be ignored.
    ret_point : bool, optional. Default: False
    verbose : bool, optional. Default: True
        Whether to print the optimization results.
    minimize_kwargs : dict, optional
        Additional arguments for the :py:func:scipy.optimize.minimize
        function.

    Returns
    -------
    resonance : list[int, int, int]
        The coefficients of the closest 3-body resonance.
    distance : float
        The distance to the closest 3-body resonance.
    """
    if bounds is None and radius is None:
        raise ValueError("Either bounds or radius must be provided.")

    if bounds is None:
        bounds = (
            max(a - radius, 1),
            a + radius,
            max(b - radius, 1),
            b + radius,
        )

    # Get the resonances in the specified area
    mmrs = mmrs_in_area(
        bounds,
        order3,
        max_coeff3,
        max_order3,
        mmr2b=False,
        verbose=verbose,
    )

    if not mmrs:
        raise ValueError("No 3-body mean-motion resonances found in the area.")

    # Initialize the minimum distance
    min_distance = np.inf
    closest_resonance = None

    # Find the closest 3-body resonance
    for resonance in mmrs:
        _, _, distance = mindist_mmr3b(a, b, resonance, **minimize_kwargs)
        if distance < min_distance:
            min_distance = distance
            closest_resonance = resonance

    if verbose:
        print(f"Closest 3-body mean-motion resonance: {closest_resonance}")
        print(f"Distance: {min_distance}")

    return closest_resonance, min_distance


def label_mmr2b(
    resonance: tuple,
    ax: plt.Axes = None,
    xaxis: bool = True,
    lims: tuple = None,
    warn: bool = True,
) -> plt.Axes:
    """Annotate a plot with the label of a 2-body resonance line.

    The label is placed where the resonance line crosses either the
    right or top axis of the plot. The resonance coefficients are
    displayed in a compact format. If the line does not cross these
    axes, a warning is printed.

    Parameters
    ----------
    resonance : tuple
        Coefficients of the resonance line (a, b) in the form a*x + b = 0.
    ax : matplotlib.axes.Axes, optional. Default: None
        The axis object on which the label will be placed.
        If not provided, the function will use the current axis.
    xaxis : bool, optional. Default: True
        Whether the resonance is along the x-axis (True) or y-axis (False).
    lims : tuple, optional. Default: None
        Custom axis limits in the format (x_min, x_max, y_min, y_max).
        If not provided, the function will use the current axis limits.
    warn : bool, optional. Default: True
        Whether to print a warning if the resonance does not cross
        the right or top axis.

    Returns
    -------
    ax : Matplotlib Axes
        The axis object with the annotations.
    """
    a, b = resonance  # Coefficients of the resonance line
    res = a / b

    # Get the current axis if not provided
    if ax is None:
        ax = plt.gca()

    # Get axis limits
    if lims:
        x_min, x_max, y_min, y_max = lims
    else:
        x_min, x_max = ax.get_xlim()
        y_min, y_max = ax.get_ylim()

    # X of y axis
    if xaxis and x_min <= res <= x_max:
        # Define label
        label = f"{a}\n{b}"  # Compact label format
        # Normalize x-coordinate for axis transform
        res_ax = (res - x_min) / (x_max - x_min)
        ax.text(res_ax, 1.02, label, transform=ax.transAxes, ha="center")
    elif not xaxis and y_min <= res <= y_max:
        # Define label
        label = f"{a} {b}"  # Multi-line label format
        # Normalize y-coordinate for axis transform
        res_ax = (res - y_min) / (y_max - y_min)
        ax.text(1.02, res_ax, label, transform=ax.transAxes, va="center")
    elif warn:
        warnings.warn(
            f"{resonance} does not cross the right or top axis.",
            stacklevel=2,
        )

    return ax


def label_mmr3b(
    resonance: tuple,
    ax: plt.Axes = None,
    lims: tuple = None,
    warn: bool = True,
) -> plt.Axes:
    """Annotate a plot with the label of a resonance line.

    The label is placed where the resonance line crosses either the
    right or top axis of the plot. The resonance coefficients are
    displayed in a compact format. If the line does not cross these
    axes, a warning is printed.

    Parameters
    ----------
    resonance : tuple
        Coefficients of the resonance line (a, b, c) in
        the form a * x + b + c / y = 0.
    ax : matplotlib.axes.Axes, optional. Default: None
        The axis object on which the label will be placed.
        If not provided, the function will use the current axis.
    lims : tuple, optional. Default: None
        Custom axis limits in the format (x_min, x_max, y_min, y_max).
        If not provided, the function will use the current axis limits.
    warn : bool, optional. Default: True
        Whether to print a warning if the resonance does not cross
        the right or top axis.

    Returns
    -------
    ax : Matplotlib Axes
        The axis object with the annotations.
    """
    a, b, c = resonance  # Coefficients of the resonance line

    # Define the inverse of the line equation
    def rinv(y):
        """Calculate x for a given y using the line equation."""
        return -(b * y + c) / (a * y)

    # Get the current axis if not provided
    if ax is None:
        ax = plt.gca()

    # Get axis limits
    if lims:
        x_min, x_max, y_min, y_max = lims
    else:
        x_min, x_max = ax.get_xlim()
        y_min, y_max = ax.get_ylim()

    # Calculate y-coordinate at x_max
    y = mmr3b(x_max, resonance)

    # Check crossing on the right axis
    if y_min <= y <= y_max:
        label = f"{a} {b} {c}"  # Compact label format
        # Normalize y-coordinate for axis transform
        y_ax = (y - y_min) / (y_max - y_min)
        ax.text(1.01, y_ax, label, transform=ax.transAxes, va="center")

    # Check crossing on the top axis
    elif x_min <= rinv(y_max) <= x_max:
        label = f"{a}\n{b}\n{c}"  # Multi-line label format
        x = rinv(y_max)
        # Normalize x-coordinate for axis transform
        x_ax = (x - x_min) / (x_max - x_min)
        ax.text(x_ax, 1.02, label, transform=ax.transAxes, ha="center")

    # If the line does not cross the right or top axis
    elif warn:
        warnings.warn(
            f"{resonance} does not cross the right or top axis.",
            stacklevel=2,
        )

    return ax


def plot_mmrs(
    bounds: Tuple[float, float, float, float] = None,
    order3: int = 0,
    max_coeff3: int = 10,
    max_order3: int = 0,
    mmr2b: bool = True,
    order2: int = 2,
    max_coeff2: int = 10,
    max_order2: int = 2,
    n_points: int = 1000,
    ax: plt.Axes = None,
    label_mmrs: bool = False,
    label_2mmrs: bool = False,
    **plot_kwargs,
):
    """Plot 3-body and 2-body mean-motion resonances in a phase-space region.

    This function plots 3-body mean-motion resonances (3P-MMRs) and optionally
    2-body mean-motion resonances (2P-MMRs) in a specified region of the phase
    space.

    Note
    ----
    The function will adjust the bounds to the axis limits if the axis object
    is provided. If the `label_mmrs` option is used, it is recommended to set
    the `xlim` and `ylim` before calling this function, or the labels may be
    placed outside the plot.

    Parameters
    ----------
    bounds : Tuple[float, float, float, float], optional. Default: None
        The limits of the region (x_min, x_max, y_min, y_max).
        If ax is provided, the bounds will be adjusted to the axis limits.
    order3 : int. Default: 0
        Exact order for 3P-MMRs.
    max_coeff3 : int, optional. Default: 10
        Calculate 3P-MMRs up to this maximum integer coefficient.
    max_order3 : int, optional. Default: 0
        Calculate 3P-MMRs up to this maximum order.
        If set to 0, only the exact order is considered (default: 10).
    mmr2b : bool, optional. Default: True
        Whether to compute 2P-MMRs.
    order2 : int, optional. Default: 2
        Exact order for 2P-MMRs.
    max_coeff2 : int, optional. Default: 10
        Calculate 2P-MMRs up to this maximum integer coefficient.
    max_order2 : int, optional. Default: 2
        Calculate 2P-MMRs up to this maximum order.
        If set to 0, only the exact order is considered.
    n_points : int, optional. Default: 500
        Number of points for the curve.
    ax : plt.Axes, optional. Default: None
        The axis object on which the plot will be drawn.
        If not provided, a new figure and axis will be created.
    label_mmrs : bool, optional. Default: False
        Whether to label the resonances on the plot.
        Recommended to set xlim and ylim before using this option,
        or the labels may be placed outside the plot. (default: False).
    label_2mmrs : bool, optional. Default: False
        Whether to label the 2-body resonances on the plot.
        Recommended to set xlim and ylim before using this option,
        or the labels may be placed outside the plot. (default: False).
    plot_kwargs : dict, optional
        Additional arguments for the plot function.

    Returns
    -------
    ax : Matplotlib Axes
        The axis object on which the plot was drawn.
    """
    # Get the current axis if not provided
    if ax is None:
        ax = plt.gca()

    if bounds is None:
        # Get the axis limits
        bounds = [
            ax.get_xlim()[0],
            ax.get_xlim()[1],
            ax.get_ylim()[0],
            ax.get_ylim()[1],
        ]

    # Get the resonances in the specified area
    mmrs = mmrs_in_area(
        bounds,
        order3,
        max_coeff3,
        max_order3,
        mmr2b,
        order2,
        max_coeff2,
        max_order2,
    )

    # Plot the resonances
    if mmr2b:
        mmr3 = mmrs[0]
        mmr2x = mmrs[1]
        mmr2y = mmrs[2]
    else:
        mmr3 = mmrs
        mmr2x = mmr2y = []

    # Plot the 2P-MMRs
    for r2x in mmr2x:
        ax.axvline(r2x[0] / r2x[1], color="k", linestyle="--")
        if label_2mmrs:
            label_mmr2b(r2x, ax, xaxis=True, warn=True)

    # Plot the 2P-MMRs
    for r2y in mmr2y:
        ax.axhline(r2y[0] / r2y[1], color="k", linestyle="--")
        if label_2mmrs:
            label_mmr2b(r2y, ax, xaxis=False, warn=True)

    # Plot the 3P-MMRs
    for r3 in mmr3:
        x = np.linspace(bounds[0], bounds[1], n_points)
        curve = mmr3b(x, r3)
        ax.plot(x, curve, **plot_kwargs)
        if label_mmrs:
            label_mmr3b(r3, ax, warn=True)

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# This file indicates that the directory should be treated as a package.

# ============================================================================
# DOCS
# ============================================================================

"""The ResoKit.load package includes tools to load planetary systems locally."""

# =============================================================================
# IMPORTS
# =============================================================================

from .load import from_binary  # noqa
from .load import from_eu  # noqa
from .load import from_nasa  # noqa

# Make the functions available at the package level

__all__ = [
    "from_binary",
    "from_eu",
    "from_nasa",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# =============================================================================
# DOCS
# =============================================================================

"""Module to query exoplanet.eu and NASA datasets with optimized structure."""

# =============================================================================
# IMPORTS
# =============================================================================

from io import BytesIO
from typing import Union

import pandas as pd

import requests

from resokit.core import (
    ResokitDataFrame,
    StaticSystem,
    df_to_resokit,
    resokit_to_system,
)
from resokit.utils.parser import DEFAULT_METADATA, assert_module_imported


try:
    from astropy.table import Table

    astropy_imported = True
except ImportError:
    astropy_imported = False

# =============================================================================
# CONSTANTS
# =============================================================================

# Query URLs for the two datasets
QUERY_URL = {
    "eu": "http://voparis-tap-planeto.obspm.fr/tap/sync",
    "nasa": "https://exoplanetarchive.ipac.caltech.edu/TAP/sync",
}


# =============================================================================
# DYNAMIC
# =============================================================================

_session_queries = {}  # This dict will store queries for the session

# =============================================================================
# FUNCTIONS
# =============================================================================


def build_query(
    source: str,
    select: str = "*",
    alias: str = "",
    conditions: str = "",
    order_by: str = "",
) -> str:
    """Construct a very simple query for the specified dataset source.

    For more information on how to write the conditions, visit:
    - EU information: http://voparis-tap-planeto.obspm.fr/__system__/dc_tables/show/tableinfo/exoplanet.epn_core
    - NASA information: https://exoplanetarchive.ipac.caltech.edu/docs/TAP/usingTAP.html#sync-query

    Parameters
    ----------
    source : str
        Data source identifier ('eu' or 'nasa').
    select : str, optional. Default: '*'.
        Columns to select in the query (default is '*').
    alias : str, optional. Default: ''.
        Optional alias for the table or columns.
    conditions : list of str, optional. Default: ''.
        Multiple (list) conditions are grouped with 'AND' clause.
        If 'OR' is needed, it must be written explicitly.
        List of conditions for WHERE clause.
    order_by : str, optional. Default: ''.
        Column name for ORDER BY clause.

    Returns
    -------
    str
        Constructed query string.
    """  # noqa: E501
    source = source.lower()  # Ensure lowercase

    # SELECT clause
    if not isinstance(select, str):
        raise ValueError("Select must be a string.")

    # Construct the query
    query = f"SELECT {select} "

    # FROM clause
    if alias:
        if not isinstance(alias, str):
            raise ValueError("Alias must be a string.")
        query += f"AS {alias} "

    # Add the source table
    query += "FROM ps" if source == "nasa" else "FROM exoplanet.epn_core"

    # WHERE clause
    if conditions:
        if isinstance(conditions, str):
            query += f" WHERE {conditions}"
        else:
            where_conditions = [f"({condition})" for condition in conditions]
            query += f" WHERE {' AND '.join(where_conditions)}"

    # ORDER BY clause
    if order_by:
        if not isinstance(order_by, str):
            raise ValueError("Order by must be a string.")
        query += f" ORDER BY {order_by}"

    return query


def execute_query(
    source: str,
    query: str,
    cache: bool = True,
    to_bytes: bool = False,
    verbose: bool = True,
    soft: bool = False,
) -> Union[bytes, pd.DataFrame]:
    """Execute a query on the specified dataset source.

    This function attempts to follow all TAPs.
    For more information on how to write the conditions, visit:
    - EU information: http://voparis-tap-planeto.obspm.fr/__system__/dc_tables/show/tableinfo/exoplanet.epn_core
    - NASA information: https://exoplanetarchive.ipac.caltech.edu/docs/TAP/usingTAP.html#sync-query

    Parameters
    ----------
    source : str
        Data source identifier ('eu' or 'nasa').
    query : str
        Query string to execute.
    cache : bool, optional. Default: True.
        Cache the query and result (dataframe) in case of repetition
        during this session.
    to_bytes: bool, optional. Default: False
        Return bytes instead of pandas dataframe.
    verbose : bool, optional. Default: True.
        Print messages
    soft : bool, optional. Default = False.
        Whether to perform a query that do not includes a
        WHERE statement. This is not recommended, as it would download
        the full databases. Use `resokit.datasets.download(...)` for this.

    Returns
    -------
    Union[bytes, pd.DataFrame]
        Resulting dataset as a pandas DataFrame, or bytes if requested.
    """  # noqa: E501
    source = source.lower()  # Ensure lowercase

    if source not in QUERY_URL:
        raise ValueError(f"Unknown data source: {source}")

    # Define the query URL
    url = QUERY_URL[source]

    # Check if WHERE is included
    if "WHERE" not in query.upper():
        if not soft:
            raise ValueError("Expected a WHERE clause in the query.")
        if verbose:
            print("Warning: No WHERE clause in the query.")

    # Define the query parameters
    params = {
        "query": query,
    }
    if source == "nasa":
        params["format"] = "csv"
    else:
        global astropy_imported
        astropy_imported = assert_module_imported(
            astropy_imported, "astropy", "(Not needed for NASA)"
        )
        params["lang"] = "ADQL"

    # Construct the full query URL (used as the cache key)
    req = requests.Request("GET", url, params=params).prepare()
    query_url = req.url  # This is the full URL with encoded params

    if query_url in _session_queries:
        if verbose:
            print("Using cached previous identic query.")
        if not isinstance(_session_queries[query_url], pd.DataFrame):
            print("Error. Last query result was not a dataframe.")
            print(" Deleting cached query. Retry if necessary.")
            raise ValueError(
                "Expected previous query result as dataframe, but"
                + f" got {type(_session_queries[query_url])} instead."
            )
        if to_bytes:
            buffer = BytesIO()
            _session_queries[query_url].to_csv(buffer)
            buffer.seek(0)
            return buffer.getvalue()
        return _session_queries[query_url]

    # Default result
    result = pd.DataFrame()

    try:  # Execute the query
        if verbose:
            print("Executing the query...")
        response = requests.get(url, params=params)
        response.raise_for_status()

        if source == "nasa":  # Parse CSV response
            result = pd.read_csv(BytesIO(response.content))
        else:  # Parse VOTable response
            result = Table.read(BytesIO(response.content)).to_pandas()

        if verbose:
            aux = " , but returned no data" if result.empty else ""
            print(f"Query executed successfully{aux}.")

    except requests.RequestException as e:
        print(f" Error querying {source} database: {e}")

    if cache:
        _session_queries.update({query_url: result})

    return result


def query_system(
    source: str,
    star_name: str = "",
    planet_name: str = "",
    default_flag: int = 1,
    controversial_flag: int = 0,
    cache: bool = True,
    verbose: bool = True,
    as_frame: bool = False,
    raw: bool = False,
) -> Union[pd.DataFrame, ResokitDataFrame, StaticSystem]:
    """Query the online dataset based on specified filters.

    Parameters
    ----------
    source : str
        Data source identifier ('eu' or 'nasa').
    star_name : str, optional. Default: ''.
        Host star or system name.
    planet_name : str, optional. Default: ''.
        Planet name.
    default_flag : int, optional. Default: 1.
        Restrict to default values in NASA dataset.
        If equal to 0, only returns non default parameter set for each planet.
        If equal to 1, only returns default parameter set for each planet.
        If None, all planet sets (default and not) are returned.
    controversial_flag : int, optional. Default: 0.
        Restrict to controversial planets in NASA dataset.
        If equal to 0, only returns confirmed planets.
        If equal to 1, only returns controversial planets.
        If None, all planets (confirmed and not) are returned.
    cache : bool, optional. Default: True.
        Cache the query and result in case of repetition
        during this session.
    verbose : bool, optional. Default: True.
        Print query information.
    as_frame : bool, optional. Default: False.
        Whether to return the dataset in a `ResokitDataFrame`
        format (True), or a `StaticSystem` format (False).
    raw : bool, optional. Default = False.
        Whether to return a pandas DataFrame of the query result.
        Overrides `as_frame` parameter.

    Returns
    -------
    data : DataFrame, ResokitDataFrame or StaticSystem
        Results of the query in a :py:class:`ResokitDataFrame`
        (if `as_resokit=True`), or :py:class:`StaticSystem`,
        or :py:class:`pandas.DataFrame` (if `raw=True`).
    """
    if not planet_name and not star_name:
        raise ValueError(
            "Either 'planet_name' or 'star_name' must be provided."
        )

    if planet_name and star_name:
        raise ValueError(
            "Only one of 'planet_name' or 'star_name' can be provided."
        )

    # Define the target or star field based on the source
    if source not in ["eu", "nasa"]:
        raise ValueError("Invalid source. Must be 'eu' or 'nasa'.")

    if source == "eu":
        field_name = "target_name" if planet_name else "star_name"
        global astropy_imported
        astropy_imported = assert_module_imported(
            astropy_imported, "astropy", "(Not needed for NASA)"
        )
    else:
        field_name = "pl_name" if planet_name else "hostname"

    filter_value = star_name or planet_name  # Get the filter value

    # Build the query
    query = build_query(source, conditions=[f"{field_name}='{filter_value}'"])

    # Add default_flag condition for NASA source
    if default_flag and source == "nasa":
        query += " AND default_flag=1"

    # Add controversial_flag condition for NASA source
    if controversial_flag is not None and source == "nasa":
        query += f" AND pl_controv_flag={controversial_flag}"

    # Print query information
    if verbose:
        print(f" Querying {source} database with query: {query}")

    # Execute query and get results
    df = execute_query(source=source, query=query)

    # Cache result
    if cache:
        _session_queries.update({query: df})

    # Get raw?
    if raw:
        return df

    # Convert to ResoKit format
    # Note: Metadata is set from default values
    meta = DEFAULT_METADATA.copy()
    meta.update({"query": query})

    reso = df_to_resokit(  # Convert to ResoKit DataFrame
        df=df,
        source=source,
        drop=False,
        copy=False,
        metadata=meta,
    )

    if not as_frame:  # Return StaticSystem
        return resokit_to_system(reso)

    return reso  # Return ResoKit DataFrame



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# This file indicates that the directory should be treated as a package.

# ============================================================================
# DOCS
# ============================================================================

"""The ResoKit.query package includes tools to make online queries."""

# =============================================================================
# IMPORTS
# =============================================================================


from .query import build_query, execute_query, query_system  # noqa

# Make the functions available at the package level

__all__ = [
    "build_query",
    "execute_query",
    "query_system",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# =============================================================================
# DOCS
# =============================================================================

"""Module with internal utility functions for the datasets package."""

# =============================================================================
# IMPORTS
# =============================================================================

import os
import re
import shutil
from datetime import datetime
from pathlib import Path
from tempfile import mkdtemp
from types import MappingProxyType
from typing import Any, Callable, List, Set, Tuple, Union
from zipfile import ZIP_DEFLATED, ZipFile

from bs4 import BeautifulSoup

from pandas import DataFrame, concat, merge, read_csv

import requests

from resokit.utils.parser import parse_to_iter


# =============================================================================
# CONSTANTS
# =============================================================================

# Path to the datasets directory
DATASETS_DIR = Path(os.path.expanduser(os.path.join("~", ".resokit_data")))

# -------------------------- EU and NASA DATASETS -----------------------------

# Name for the ZIP archive
DATASET_ZIPNAMES = {"eu": "exoplanet_eu.zip", "nasa": "nasa_exoplanets.zip"}

# Filenames and URLs for the datasets
DATASET_FILENAMES = {"eu": "exoplanet_eu.csv", "nasa": "nasa.csv"}
DATASET_URLS = {
    "eu": "https://exoplanet.eu/catalog/csv/",
    "nasa": "https://exoplanetarchive.ipac.caltech.edu/TAP/sync?"
    + "query=select+*+from+ps&format=csv",
}

# Index columns for each dataset
INDEX_COLUMNS = {"eu": ["name", "star_name"], "nasa": ["pl_name", "hostname"]}

# EU dtypes
EU_MAPPING = {
    "name": "object",
    "planet_status": "object",
    "mass": "float64",
    "mass_error_min": "float64",
    "mass_error_max": "float64",
    "mass_sini": "float64",
    "mass_sini_error_min": "float64",
    "mass_sini_error_max": "float64",
    "radius": "float64",
    "radius_error_min": "float64",
    "radius_error_max": "float64",
    "orbital_period": "float64",
    "orbital_period_error_min": "float64",
    "orbital_period_error_max": "float64",
    "semi_major_axis": "float64",
    "semi_major_axis_error_min": "float64",
    "semi_major_axis_error_max": "float64",
    "eccentricity": "float64",
    "eccentricity_error_min": "float64",
    "eccentricity_error_max": "float64",
    "inclination": "float64",
    "inclination_error_min": "float64",
    "inclination_error_max": "float64",
    "angular_distance": "float64",
    "discovered": "float64",
    "updated": "object",
    "omega": "float64",
    "omega_error_min": "float64",
    "omega_error_max": "float64",
    "tperi": "float64",
    "tperi_error_min": "float64",
    "tperi_error_max": "float64",
    "tconj": "float64",
    "tconj_error_min": "float64",
    "tconj_error_max": "float64",
    "tzero_tr": "float64",
    "tzero_tr_error_min": "float64",
    "tzero_tr_error_max": "float64",
    "tzero_tr_sec": "float64",
    "tzero_tr_sec_error_min": "float64",
    "tzero_tr_sec_error_max": "float64",
    "lambda_angle": "float64",
    "lambda_angle_error_min": "float64",
    "lambda_angle_error_max": "float64",
    "impact_parameter": "float64",
    "impact_parameter_error_min": "float64",
    "impact_parameter_error_max": "float64",
    "tzero_vr": "float64",
    "tzero_vr_error_min": "float64",
    "tzero_vr_error_max": "float64",
    "k": "float64",
    "k_error_min": "float64",
    "k_error_max": "float64",
    "temp_calculated": "float64",
    "temp_calculated_error_min": "float64",
    "temp_calculated_error_max": "float64",
    "temp_measured": "float64",
    "hot_point_lon": "float64",
    "geometric_albedo": "float64",
    "geometric_albedo_error_min": "float64",
    "geometric_albedo_error_max": "float64",
    "log_g": "float64",
    "publication": "object",
    "detection_type": "object",
    "mass_measurement_type": "object",
    "radius_measurement_type": "object",
    "alternate_names": "object",
    "molecules": "object",
    "star_name": "object",
    "ra": "float64",
    "dec": "float64",
    "mag_v": "float64",
    "mag_i": "float64",
    "mag_j": "float64",
    "mag_h": "float64",
    "mag_k": "float64",
    "star_distance": "float64",
    "star_distance_error_min": "float64",
    "star_distance_error_max": "float64",
    "star_metallicity": "float64",
    "star_metallicity_error_min": "float64",
    "star_metallicity_error_max": "float64",
    "star_mass": "float64",
    "star_mass_error_min": "float64",
    "star_mass_error_max": "float64",
    "star_radius": "float64",
    "star_radius_error_min": "float64",
    "star_radius_error_max": "float64",
    "star_sp_type": "object",
    "star_age": "float64",
    "star_age_error_min": "float64",
    "star_age_error_max": "float64",
    "star_teff": "float64",
    "star_teff_error_min": "float64",
    "star_teff_error_max": "float64",
    "star_detected_disc": "object",
    "star_magnetic_field": "object",
    "star_alternate_names": "object",
}

# Nasa dtypes
NASA_MAPPING = {
    "pl_name": "object",
    "pl_letter": "object",
    "hostname": "object",
    "hd_name": "object",
    "hip_name": "object",
    "tic_id": "object",
    "gaia_id": "object",
    "default_flag": "Int64",
    "pl_refname": "object",
    "sy_refname": "object",
    "disc_pubdate": "object",
    "disc_year": "Int64",
    "discoverymethod": "object",
    "disc_locale": "object",
    "disc_facility": "object",
    "disc_instrument": "object",
    "disc_telescope": "object",
    "disc_refname": "object",
    "ra": "float64",
    "rastr": "object",
    "dec": "float64",
    "decstr": "object",
    "glon": "float64",
    "glat": "float64",
    "elon": "float64",
    "elat": "float64",
    "pl_orbper": "float64",
    "pl_orbpererr1": "float64",
    "pl_orbpererr2": "float64",
    "pl_orbperlim": "float64",
    "pl_orbperstr": "object",
    "pl_orblpererr1": "float64",
    "pl_orblper": "float64",
    "pl_orblpererr2": "float64",
    "pl_orblperlim": "float64",
    "pl_orblperstr": "object",
    "pl_orbsmax": "float64",
    "pl_orbsmaxerr1": "float64",
    "pl_orbsmaxerr2": "float64",
    "pl_orbsmaxlim": "float64",
    "pl_orbsmaxstr": "object",
    "pl_orbincl": "float64",
    "pl_orbinclerr1": "float64",
    "pl_orbinclerr2": "float64",
    "pl_orbincllim": "float64",
    "pl_orbinclstr": "object",
    "pl_orbtper": "float64",
    "pl_orbtpererr1": "float64",
    "pl_orbtpererr2": "float64",
    "pl_orbtperlim": "float64",
    "pl_orbtperstr": "object",
    "pl_orbeccen": "float64",
    "pl_orbeccenerr1": "float64",
    "pl_orbeccenerr2": "float64",
    "pl_orbeccenlim": "float64",
    "pl_orbeccenstr": "object",
    "pl_eqt": "float64",
    "pl_eqterr1": "float64",
    "pl_eqterr2": "float64",
    "pl_eqtlim": "float64",
    "pl_eqtstr": "object",
    "pl_occdep": "float64",
    "pl_occdeperr1": "float64",
    "pl_occdeperr2": "float64",
    "pl_occdeplim": "float64",
    "pl_occdepstr": "object",
    "pl_insol": "float64",
    "pl_insolerr1": "float64",
    "pl_insolerr2": "float64",
    "pl_insollim": "float64",
    "pl_insolstr": "object",
    "pl_dens": "float64",
    "pl_denserr1": "float64",
    "pl_denserr2": "float64",
    "pl_denslim": "float64",
    "pl_densstr": "object",
    "pl_trandep": "float64",
    "pl_trandeperr1": "float64",
    "pl_trandeperr2": "float64",
    "pl_trandeplim": "float64",
    "pl_trandepstr": "object",
    "pl_tranmid": "float64",
    "pl_tranmiderr1": "float64",
    "pl_tranmiderr2": "float64",
    "pl_tranmidlim": "float64",
    "pl_tranmidstr": "object",
    "pl_trandur": "float64",
    "pl_trandurerr1": "float64",
    "pl_trandurerr2": "float64",
    "pl_trandurlim": "float64",
    "pl_trandurstr": "object",
    "sy_kmagstr": "object",
    "sy_umag": "float64",
    "sy_umagerr1": "float64",
    "sy_umagerr2": "float64",
    "sy_umagstr": "object",
    "sy_rmag": "float64",
    "sy_rmagerr1": "float64",
    "sy_rmagerr2": "float64",
    "sy_rmagstr": "object",
    "sy_imag": "float64",
    "sy_imagerr1": "float64",
    "sy_imagerr2": "float64",
    "sy_imagstr": "object",
    "sy_zmag": "float64",
    "sy_zmagerr1": "float64",
    "sy_zmagerr2": "float64",
    "sy_zmagstr": "object",
    "sy_w1mag": "float64",
    "sy_w1magerr1": "float64",
    "sy_w1magerr2": "float64",
    "sy_w1magstr": "object",
    "sy_w2mag": "float64",
    "sy_w2magerr1": "float64",
    "sy_w2magerr2": "float64",
    "sy_w2magstr": "object",
    "sy_w3mag": "float64",
    "sy_w3magerr1": "float64",
    "sy_w3magerr2": "float64",
    "sy_w3magstr": "object",
    "sy_w4mag": "float64",
    "sy_w4magerr1": "float64",
    "sy_w4magerr2": "float64",
    "sy_w4magstr": "object",
    "sy_gmag": "float64",
    "sy_gmagerr1": "float64",
    "sy_gmagerr2": "float64",
    "sy_gmagstr": "object",
    "sy_gaiamag": "float64",
    "sy_gaiamagerr1": "float64",
    "sy_gaiamagerr2": "float64",
    "sy_gaiamagstr": "object",
    "sy_tmag": "float64",
    "sy_tmagerr1": "float64",
    "sy_tmagerr2": "float64",
    "sy_tmagstr": "object",
    "pl_controv_flag": "Int64",
    "pl_tsystemref": "object",
    "st_metratio": "object",
    "st_spectype": "object",
    "sy_kepmag": "float64",
    "sy_kepmagerr1": "float64",
    "sy_kepmagerr2": "float64",
    "sy_kepmagstr": "float64",
    "st_rotp": "float64",
    "st_rotperr1": "float64",
    "st_rotperr2": "float64",
    "st_rotplim": "float64",
    "st_rotpstr": "object",
    "pl_projobliq": "float64",
    "pl_projobliqerr1": "float64",
    "pl_projobliqerr2": "float64",
    "pl_projobliqlim": "float64",
    "pl_projobliqstr": "object",
    "x": "float64",
    "y": "float64",
    "z": "float64",
    "htm20": "Int64",
    "pl_rvamp": "float64",
    "pl_rvamperr1": "float64",
    "pl_rvamperr2": "float64",
    "pl_rvamplim": "float64",
    "pl_rvampstr": "object",
    "pl_radj": "float64",
    "pl_radjerr1": "float64",
    "pl_radjerr2": "float64",
    "pl_radjlim": "float64",
    "pl_radjstr": "object",
    "pl_rade": "float64",
    "pl_radeerr1": "float64",
    "pl_radeerr2": "float64",
    "pl_radelim": "float64",
    "pl_radestr": "object",
    "pl_ratror": "float64",
    "pl_ratrorerr1": "float64",
    "pl_ratrorerr2": "float64",
    "pl_ratrorlim": "float64",
    "pl_ratrorstr": "object",
    "pl_ratdor": "float64",
    "pl_trueobliq": "float64",
    "pl_trueobliqerr1": "float64",
    "pl_trueobliqerr2": "float64",
    "pl_trueobliqlim": "float64",
    "pl_trueobliqstr": "object",
    "sy_icmag": "float64",
    "sy_icmagerr1": "float64",
    "sy_icmagerr2": "float64",
    "sy_icmagstr": "object",
    "rowupdate": "object",
    "pl_pubdate": "object",
    "st_refname": "object",
    "releasedate": "object",
    "dkin_flag": "Int64",
    "pl_ratdorerr1": "float64",
    "pl_ratdorerr2": "float64",
    "pl_ratdorlim": "float64",
    "pl_ratdorstr": "object",
    "pl_imppar": "float64",
    "pl_impparerr1": "float64",
    "pl_impparerr2": "float64",
    "pl_impparlim": "float64",
    "pl_impparstr": "object",
    "pl_cmassj": "float64",
    "pl_cmassjerr1": "float64",
    "pl_cmassjerr2": "float64",
    "pl_cmassjlim": "float64",
    "pl_cmassjstr": "object",
    "pl_cmasse": "float64",
    "pl_cmasseerr1": "float64",
    "pl_cmasseerr2": "float64",
    "pl_cmasselim": "float64",
    "pl_cmassestr": "object",
    "pl_massj": "float64",
    "pl_massjerr1": "float64",
    "pl_massjerr2": "float64",
    "pl_massjlim": "float64",
    "pl_massjstr": "object",
    "pl_masse": "float64",
    "pl_masseerr1": "float64",
    "pl_masseerr2": "float64",
    "pl_masselim": "float64",
    "pl_massestr": "object",
    "pl_bmassj": "float64",
    "pl_bmassjerr1": "float64",
    "pl_bmassjerr2": "float64",
    "pl_bmassjlim": "float64",
    "pl_bmassjstr": "object",
    "pl_bmasse": "float64",
    "pl_bmasseerr1": "float64",
    "pl_bmasseerr2": "float64",
    "pl_bmasselim": "float64",
    "pl_bmassestr": "object",
    "pl_bmassprov": "object",
    "pl_msinij": "float64",
    "pl_msinijerr1": "float64",
    "pl_msinijerr2": "float64",
    "pl_msinijlim": "float64",
    "pl_msinijstr": "object",
    "pl_msinie": "float64",
    "pl_msinieerr1": "float64",
    "pl_msinieerr2": "float64",
    "pl_msinielim": "float64",
    "pl_msiniestr": "object",
    "st_teff": "float64",
    "st_tefferr1": "float64",
    "st_tefferr2": "float64",
    "st_tefflim": "float64",
    "st_teffstr": "object",
    "st_met": "float64",
    "st_meterr1": "float64",
    "st_meterr2": "float64",
    "st_metlim": "float64",
    "st_metstr": "object",
    "st_radv": "float64",
    "st_radverr1": "float64",
    "st_radverr2": "float64",
    "st_radvlim": "float64",
    "st_radvstr": "object",
    "st_vsin": "float64",
    "st_vsinerr1": "float64",
    "st_vsinerr2": "float64",
    "st_vsinlim": "float64",
    "st_vsinstr": "object",
    "st_lum": "float64",
    "st_lumerr1": "float64",
    "st_lumerr2": "float64",
    "st_lumlim": "float64",
    "st_lumstr": "object",
    "st_logg": "float64",
    "st_loggerr1": "float64",
    "st_loggerr2": "float64",
    "st_logglim": "float64",
    "st_loggstr": "object",
    "st_age": "float64",
    "st_ageerr1": "float64",
    "st_ageerr2": "float64",
    "st_agelim": "float64",
    "st_agestr": "object",
    "st_mass": "float64",
    "st_masserr1": "float64",
    "st_masserr2": "float64",
    "st_masslim": "float64",
    "st_massstr": "object",
    "st_dens": "float64",
    "st_denserr1": "float64",
    "st_denserr2": "float64",
    "st_denslim": "float64",
    "st_densstr": "object",
    "st_rad": "float64",
    "st_raderr1": "float64",
    "st_raderr2": "float64",
    "st_radlim": "float64",
    "st_radstr": "object",
    "ttv_flag": "Int64",
    "ptv_flag": "Int64",
    "tran_flag": "Int64",
    "rv_flag": "Int64",
    "ast_flag": "Int64",
    "obm_flag": "Int64",
    "micro_flag": "Int64",
    "etv_flag": "Int64",
    "ima_flag": "Int64",
    "pul_flag": "Int64",
    "soltype": "object",
    "sy_snum": "Int64",
    "sy_pnum": "Int64",
    "sy_mnum": "Int64",
    "cb_flag": "Int64",
    "st_nphot": "Int64",
    "st_nrvc": "Int64",
    "st_nspec": "Int64",
    "pl_nespec": "Int64",
    "pl_ntranspec": "Int64",
    "pl_ndispec": "Int64",
    "pl_nnotes": "Int64",
    "sy_pm": "float64",
    "sy_pmerr1": "float64",
    "sy_pmerr2": "float64",
    "sy_pmstr": "object",
    "sy_pmra": "float64",
    "sy_pmraerr1": "float64",
    "sy_pmraerr2": "float64",
    "sy_pmrastr": "object",
    "sy_pmdec": "float64",
    "sy_pmdecerr1": "float64",
    "sy_pmdecerr2": "float64",
    "sy_pmdecstr": "object",
    "sy_plx": "float64",
    "sy_plxerr1": "float64",
    "sy_plxerr2": "float64",
    "sy_plxstr": "object",
    "sy_dist": "float64",
    "sy_disterr1": "float64",
    "sy_disterr2": "float64",
    "sy_diststr": "object",
    "sy_bmag": "float64",
    "sy_bmagerr1": "float64",
    "sy_bmagerr2": "float64",
    "sy_bmagstr": "object",
    "sy_vmag": "float64",
    "sy_vmagerr1": "float64",
    "sy_vmagerr2": "float64",
    "sy_vmagstr": "object",
    "sy_jmag": "float64",
    "sy_jmagerr1": "float64",
    "sy_jmagerr2": "float64",
    "sy_jmagstr": "object",
    "sy_hmag": "float64",
    "sy_hmagerr1": "float64",
    "sy_hmagerr2": "float64",
    "sy_hmagstr": "object",
    "sy_kmag": "float64",
    "sy_kmagerr1": "float64",
    "sy_kmagerr2": "float64",
}

# Mapping of dataset names to their respective dtypes
DATASET_DTYPES = MappingProxyType({"eu": EU_MAPPING, "nasa": NASA_MAPPING})

# URL to query the dataset length
QUERY_LENGTH_URL = {
    "eu": "https://exoplanet.eu/home/",
    "nasa": "https://exoplanetarchive.ipac.caltech.edu/index.html",
}

# --------------------------- BINARY SYSTEMS DATASETS --------------------------

# No ZIP here, only txt

# Filenames and URLs for the binaries datasets
BINARIES_FILENAMES = {"p": "plan_circ.txt", "s": "plan_bin500aun.txt"}
BINARIES_URLS = {
    "p": "https://lesia.obspm.fr/perso/philippe-thebault/plan_circ.txt",
    "s": "https://lesia.obspm.fr/perso/philippe-thebault/plan_bin500aun.txt",
}

# Columns of the binaries datasets
BINARIES_COLUMNS = [
    "star0_name",
    "alternate_name",
    "star0_mass",
    "star1_mass",
    "dist",
    "disc_method",
    "a",
    "e",
    "nplanets",
    "planet_a",
    "planet_e",
    "planet_mass",
    "planet_HW_crit",
    "imut",
]

# Default key columns to use to match old and new rows
DEFAULT_KEY_COLS = {
    "nasa": [
        "pl_name",
        "pl_letter",
        "hostname",
        "default_flag",
        "pl_controv_flag",
        "releasedate",
        "pl_refname",
    ],
    "eu": [
        "name",
        "updated",
        "star_name",
    ],
}

# =============================================================================
# FUNCTIONS
# =============================================================================


def resolve_paths(
    to_file: Union[bool, str, Path],
    to_zip: Union[bool, str, Path],
    dir_path: Union[bool, str, Path, None],
    default_file: str,
    default_zip: str,
    default_dir: Path,
) -> Tuple[Set[Path], Set[Path], Set[Path]]:
    """
    Normalize to_file, to_zip, and dir_path into full output file and zip paths.

    Note
    ----
    This function is not intended to by explicitly executed by the user.

    Parameters
    ----------
        to_file: File name or path (True = use default file name in dir_path).
        to_zip: Zip name or path (True = use default zip name in dir_path).
        dir_path: Base directory (True = use default_dir).
        default_file: Default file name if to_file is True.
        default_zip: Default zip name if to_zip is True.
        default_dir: Default directory if dir_path is True.

    Returns
    -------
        Tuple containing:
            - Set of base directories (Path): physical directories containing
               files/zips
            - Set of resolved file paths (Path)
            - Set of resolved zip output paths (Path to file inside zip)
    """
    if to_file is True and to_zip is True and dir_path is True:
        return (
            set([default_dir]),
            set(),
            set([default_dir / default_zip / default_file]),
        )

    # Short-circuit: if to_file is explicitly False, return all empty
    if to_file is False:
        return set(), set(), set()

    def parse_path_input(value, default_name) -> Tuple[Path, Path]:
        if value is True:
            return None, Path(default_name)
        elif isinstance(value, (str, Path)):
            value = Path(value)
            if value.is_absolute() or value.parent != Path("."):
                return value.parent.resolve(), value.name
            else:
                return None, value.name
        elif value in [False, None]:
            return None, None
        else:
            raise ValueError(f"Invalid path input: {value}")

    # Parse inputs
    file_dir, file_name = parse_path_input(to_file, default_file)
    zip_dir, zip_name = parse_path_input(to_zip, default_zip)

    # Resolve base directory
    if dir_path is True:
        base_dir = default_dir.resolve()
    elif isinstance(dir_path, (str, Path)):
        base_dir = Path(dir_path).resolve()
    else:
        base_dir = None

    # If zip_dir is not provided and dir_path is, assume zip is in base_dir
    if zip_dir is None and zip_name is not None and base_dir is not None:
        zip_dir = base_dir

    fpaths = set()
    zfpaths = set()
    base_paths = set()

    # Resolve file path
    if file_name:
        # If user gave no file_dir and zip_dir == base_dir,
        # skip adding file outside zip
        if file_dir is None and zip_dir == base_dir and base_dir is not None:
            pass  # Do not add to fpaths
        elif file_dir is None and base_dir is None and zip_name is not None:
            pass
        elif file_dir and (file_dir.name.endswith(".zip")):
            full_zfile_path = file_dir / file_name
            zfpaths.add(full_zfile_path)
        else:
            if file_dir:
                full_file_path = file_dir / file_name
            elif base_dir and not (zip_dir == base_dir):
                full_file_path = base_dir / file_name
            else:
                full_file_path = Path(file_name)
            fpaths.add(full_file_path)
            base_paths.add(full_file_path.parent)

    # Resolve zip file path (and zip-internal file path)
    if zip_name:
        if zip_dir:
            zip_file_path = zip_dir / zip_name
        elif base_dir:
            zip_file_path = base_dir / zip_name
        else:
            zip_file_path = Path(zip_name)
        if file_name:
            full_zip_entry = zip_file_path / file_name
        else:
            full_zip_entry = zip_file_path
        zfpaths.add(full_zip_entry)
        base_paths.add(
            zip_file_path.parent
        )  # <-- Only the containing folder, not the zip itself

    return base_paths, fpaths, zfpaths


def check_file_age(
    file_path: Union[str, Path],
    zip_path: Union[str, Path, None],
    verbose: bool = True,
) -> int:
    """Check the dataset file's age in days.

    Note
    ----
    This function is not intended to by explicitly executed by the user.

    Parameters
    ----------
    file_path : str or Path, optional. Default: False.
        Path to the file.
    zip_path : str or Path or None, optional. Default: False.
        Path to the ZIP archive.
    verbose : bool, optional. Default: True.
        If `True`, prints messages about the process.

    Returns
    -------
    age : int
        Age of the file in days.
    """
    file_path = Path(file_path)  # Convert to Path object
    # Get the file's last modified date
    if zip_path:
        file_name = Path(file_path).name
        with ZipFile(zip_path, "r") as zipf:  # Open the ZIP archive
            # Check if the file is in the ZIP archive
            if file_name not in zipf.namelist():
                raise FileNotFoundError(
                    f"File {file_name} not found in {zip_path}."
                )
            # Get the file's last modified date
            date_info = zipf.getinfo(file_name).date_time
            creation = datetime(*date_info)
    else:
        creation = datetime.fromtimestamp(file_path.stat().st_mtime)

    # Calculate age in days
    age = (datetime.now() - creation).days

    if verbose:
        print(f" Last modified: {creation} ({age} days ago).")

    return age


def load_from_zip(
    zip_path: Path,
    file_name: str,
    source: Union[str, None] = None,
    skip_rows: Union[int, Callable, None] = None,
    usecols: Union[list, Callable, None] = None,
    verbose: bool = True,
    custom_load: Union[Callable, None] = None,
) -> Union[DataFrame, Any]:
    """Load the dataset from a ZIP archive.

    Reads the dataset from a ZIP archive and returns it as a pandas DataFrame.

    Note
    ----
    This function is not intended to by explicitly executed by the user.

    Parameters
    ----------
    zip_path : Path
        Full path to the ZIP archive.
    file_name : str
        Name of the file to load from the ZIP archive.
    source : str, optional
        Identifier for the data source ('eu' or 'nasa').
        Used to define the dtypes of the dataset.
        If not 'eu' or 'nasa', no dtypes are defined.
    skip_rows : int, optional
        Number of rows to skip.
    usecols : list, optional
        Columns to load.
    verbose : bool, optional
        If `True`, prints messages about the process.
    custom_load : callable, optional
        Custom function to load the data from the ZIP archive.
        If provided, it is used instead of the default read_csv.
        It neglects the skip_rows, usecols, and dtypes parameters.

    Returns
    -------
    Union[pd.DataFrame, any]
        data : pd.DataFrame
            Loaded dataset as a pandas DataFrame.
        data : any
            Loaded data from the ZIP archive, if custom_load is provided.
    """
    # Check if the zip exists
    if not zip_path.exists():
        raise FileNotFoundError(f"ZIP archive {zip_path} not found.")

    # Load the dataset from the ZIP archive
    with ZipFile(zip_path, "r") as zipf:  # Open the ZIP archive
        if verbose:  # Print message if verbose
            print(f"  Reading {file_name} " + f"directly from {zip_path}...")
        # Load directly from the .zip
        dtypes = None if source is None else DATASET_DTYPES.get(source, None)
        skip_rows = 0 if skip_rows is None else skip_rows
        with zipf.open(file_name) as file:
            if custom_load is not None:
                return custom_load(file)
            data = read_csv(
                file,
                header=0,
                skiprows=skip_rows,
                usecols=usecols,
                dtype=dtypes,
            )

    return data


def remove_from_zip(zipfname: str, *filenames: str, verbose: bool = False):
    """Remove files from a zip archive.

    This function removes files from a zip archive without extracting it.
    It is unefficient (especially for large archives) because it decompresses
    and recompresses the whole archive.

    Note
    ----
    This function is not intended to by explicitly executed by the user.

    Parameters
    ----------
    zipfname : str
        Path to the zip archive.
    filenames : str
        Names of the files to remove from the archive.
    verbose : bool, optional
        If True, print messages about the process.

    Returns
    -------
    None
    """
    # Check if any of the files to remove is in the archive
    has_files = False
    with ZipFile(zipfname, "r") as zipread:
        for filename in filenames:
            if filename in zipread.namelist():
                has_files = True
                break
    if not has_files:
        return

    # Create a temporary directory
    tempdir = mkdtemp()
    try:
        # Create a new zip archive
        tempname = os.path.join(tempdir, "new.zip")
        # Read the original archive
        with ZipFile(zipfname, "r") as zipread:
            # Write the new archive
            with ZipFile(tempname, "w", compression=ZIP_DEFLATED) as zipwrite:
                # Copy all files except the ones to remove
                for item in zipread.infolist():
                    if item.filename not in filenames:
                        data = zipread.read(item.filename)
                        zipwrite.writestr(item, data)
        # Replace the original archive with the new one
        shutil.move(tempname, zipfname)
        if verbose:
            print(f"Removed files: {', '.join(filenames)} from {zipfname}")
    finally:
        # Remove the temporary directory
        shutil.rmtree(tempdir)


def _get_request_with_refresh(url: str, verbose: bool = True, **kwargs):
    """Request a URL and automatically follows HTML meta refresh redirects."""
    response = requests.get(url, **kwargs)

    # Regex for meta refresh
    meta_match = re.search(
        r'<meta\s+[^>]*http-equiv=["\']?Refresh["\']?[^>]'
        r'*content=["\']?\s*\d+\s*;\s*URL=([^"\'>\s]+)',
        response.text,
        flags=re.IGNORECASE,
    )

    if meta_match is not None:
        redirect_url = meta_match.group(1).strip()
        # If the redirect is relative, build absolute
        if not redirect_url.lower().startswith("http"):
            redirect_url = requests.compat.urljoin(response.url, redirect_url)

        if verbose:
            print(f"Following meta refresh to: {redirect_url}")
        return requests.get(redirect_url, **kwargs)

    return response


def request_dataset(
    url: str,
    verbose: bool = True,
    chunk_size: int = 1024,
    print_size: float = 0.15,
) -> bytes:
    """Download the data from a specified URL.

    Note
    ----
    This function is not intended to by explicitly executed by the user.

    Parameters
    ----------
    url : str
        URL to download the data from.
    verbose : bool, optional
        If True, print messages about the download process.
    chunk_size : int, optional
        Size of the chunks (in bytes) to download the data.
    print_size : float, optional. Default is 0.15
        Update frequency (in MB, or KB if <=0.001) for the
        download progress bar.
        Used only if verbose is True. Useful for large downloads,
        to avoid IO overhead, especially in Jupyter notebooks.

    Returns
    -------
    content : bytes
        The downloaded data.
    """
    # Print message
    if verbose:
        print(f"Downloading data from {url}...")

    # Check if the print size is in MB or KB
    print_unit = "MB"
    bytes_unit = 1e6
    if print_size <= 0.001:
        print_size *= 1e3
        print_unit = "KB"
        bytes_unit = 1e3

    # Check if Jupyter notebook is running
    if is_notebook() and verbose:
        print(f" Note: Progress is shown at every {print_size} {print_unit}")

    # Send a GET request with streaming enabled
    response = _get_request_with_refresh(url, verbose=verbose, stream=True)

    # Check for errors
    response.raise_for_status()

    # Initialize the downloaded data as a byte array
    downloaded_data = bytearray()

    # Initialize the number of downloaded bytes
    downloaded_size = 0

    # Iterate over the response content
    increment = 0.0
    for chunk in response.iter_content(chunk_size=chunk_size):
        if chunk:  # Filter out keep-alive new chunks
            downloaded_data.extend(chunk)  # Append the chunk to the data
            downloaded_size += len(
                chunk
            )  # Update the downloaded size (in bytes)

            # Print the download progress (in MB or KB)
            if verbose and downloaded_size >= bytes_unit * increment:
                print(
                    f" Downloaded: {downloaded_size / bytes_unit:.3f} "
                    + f"{print_unit}",
                    end="\r",
                )
                increment += print_size
    # Check printed 0.0 MB
    if verbose and downloaded_size <= 1e3 and print_unit == "MB":
        print(
            f" Downloaded: {downloaded_size / 1e3:.3f} KB",
            end="\r",
        )

    return bytes(
        downloaded_data
    )  # Return the downloaded data as inmutable bytes


def is_notebook() -> bool:
    """Check if the code is running in a Jupyter notebook."""
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def __parse_date(date_str: str, soft: bool = True) -> Union[datetime, bool]:
    """Try different date formats to robustly parse the last update date.

    Parameters
    ----------
    date_str : str
        The date string to parse.
        soft : bool, optional. Default: True.
        If True, return False if parsing fails instead of raising an error.

    Returns
    -------
    Union[datetime, bool]
        Parsed date as a datetime object if successful,
        or False if parsing fails and soft is True.
    """
    # Full & abbreviated months
    date_formats = ["%B %d, %Y", "%b %d, %Y", "%b. %d, %Y", "%m/%d/%Y"]

    for fmt in date_formats:
        try:
            return datetime.strptime(date_str, fmt)
        except ValueError:
            continue  # Try the next format

    # If all fail
    if soft:
        return False
    raise ValueError(f"Could not parse date: {date_str}")


def check_outdated_dataset(
    source: str, verbose: bool = True
) -> Tuple[int, Union[int, None]]:
    """Web scrap the length (count) of the dataset from the specified source.

    Note
    ----
    This function is not intended to by explicitly executed by the user.

    Parameters
    ----------
    source : str
        Data source identifier ('eu' or 'nasa').
    verbose : bool, optional. Default: True.
        Print query information.

    Returns
    -------
    length : int
        Amount of entries (rows) in the dataset.
    last_update : datetime
        Date of the last update of the dataset.
    If no match is found, -1 is returned for length and None for last_update.
    """
    source = source.lower()  # Ensure lowercase

    # Message
    if verbose:
        print(f"Checking online dataset from {source}...")

    # Call subfunction
    if source == "eu":
        length_str, date_str = _check_outdated_eu(verbose=verbose)
    elif source == "nasa":
        length_str, date_str = _check_outdated_nasa(verbose=verbose)
    else:
        raise ValueError("Invalid source. Must be 'eu' or 'nasa'.")

    # Parse the planets
    try:
        # Remove comma if present
        length = int(str(length_str).replace(",", "").replace(".", ""))
    except ValueError:
        if verbose:
            print(
                " Could not parse the amount of planets in "
                + f"online {source} database."
            )
        length = -1

    # Parse the Date
    date = __parse_date(str(date_str), soft=True)
    if date is False:
        if verbose:
            print(" Could not parse the last update date.")
        date = None
        last_update = None
    else:
        last_update = (datetime.now() - date).days

    # Message
    if verbose:
        if length > 0:
            print(f" Number of planets in online dataset: {length}")
        if date:
            print(f" Last online update: {date} ({last_update} days ago)")

    return length, last_update


def _check_outdated_eu(
    verbose: bool = True,
) -> Tuple[str, Union[str, None]]:
    """Web scrap the length of the exoplanet.eu dataset.

    Parameters
    ----------
    verbose : bool, optional. Default: True.
        Print query information.

    Returns
    -------
    length : str
        Amount of entries (rows) in the dataset.
        If no match is found, -1 is returned.
    last_update : str
        Date of the last update of the dataset.
        If no match is found, None is returned.
    """
    try:
        # Fetch the webpage content
        response = requests.get(QUERY_LENGTH_URL["eu"], timeout=10)
        response.raise_for_status()  # Raise an error for HTTP issues
        soup = BeautifulSoup(response.text, "html.parser")

        # Find the text containing "Last update"
        text = None
        for p in soup.find_all("p"):
            if "Last update" in p.text:
                text = p.text.strip()
                break

        if text is None:
            if verbose:
                print(" No 'Last update' text found on the webpage.")
            return "-1", None

        # Extract the date and number of planets using regex
        aux = (
            r"Last\s+update[d]?\s*:\s*([\w\.]+\s+\d{1,2},\s+\d{4})"
            + r"\s+currently\s+([\d,]+)\s+planets"
        )
        match = re.search(aux, text)

        if not match:
            if verbose:
                print(" No match found in the extracted text.")
            return "-1", None

        # Parse extracted values
        last_update, length = match.groups()

        return str(length), last_update

    except requests.exceptions.RequestException as e:
        if verbose:
            print(f"Error fetching the webpage: {e}")

    return "-1", None


def _check_outdated_nasa(
    verbose: bool = True,
) -> Tuple[Union[str, int], Union[str, None]]:
    """Web scrap the length of the NASA dataset.

    Parameters
    ----------
    verbose : bool, optional. Default: True.
        Print query information.

    Returns
    -------
    length : str
        Amount of entries (rows) in the dataset.
        If no match is found, -1 is returned.
    last_update : str
        Date of the last update of the dataset.
        If no match is found, None is returned.
    """
    try:
        # Fetch the webpage content
        response = requests.get(QUERY_LENGTH_URL["nasa"])
        response.raise_for_status()  # Raise an error for HTTP issues
        soup = BeautifulSoup(response.text, "html.parser")

        # Extract the number of confirmed planets
        planet_count_div = soup.find("div", class_="stat")
        if planet_count_div:
            length = (
                planet_count_div.text.strip().replace(",", "").replace(".", "")
            )
        else:
            if verbose:
                print(" No match found in the extracted text.")
            length = -1

        # Extract the last update date
        date_div = soup.find("div", class_="date")
        if date_div:
            last_update = date_div.text.strip()
        else:
            if verbose:
                print(" No match found in the extracted text.")
            last_update = None

        return str(length), last_update

    except requests.exceptions.RequestException as e:
        if verbose:
            print(f"Error fetching the webpage: {e}")

    return "-1", None


def check_outdated_binary(source: str, verbose: bool = True) -> int:
    """Web scrap the length (count) of the file from the specified source.

    This function is kind of dumb, because it downloads the file (although
    if does not parse it, just count the lines).

    Note
    ----
    This function is not intended to by explicitly executed by the user.

    Parameters
    ----------
    source : str
        Data source identifier ('p' or 's').
    verbose : bool, optional. Default: True.
        Print query information.

    Returns
    -------
    length : int
        Amount of entries (rows) in the file.
        If no match is found, -1 is returned.
    """
    source = source.lower()  # Ensure lowercase
    if source not in BINARIES_URLS:
        raise ValueError(f"Invalid source: {source}. Must be 'p' or 's'.")

    # Message
    if verbose:
        print(f"Checking online dataset from {source}-type binaries...")

    # Call subfunction
    url = BINARIES_URLS[source]
    length = 0
    try:
        response = requests.get(url, stream=True)
        response.raise_for_status()
        for _ in response.iter_lines():
            length += 1
    except requests.RequestException as e:
        if verbose:
            print(f"Error: {e}")
            print(
                " Could not parse the amount of lines in "
                + f"online {source}-type binaries file."
            )
        length = -1

    # Message
    if verbose:
        if length > 0:
            print(
                " Number of lines (including header) "
                + f"in online dataset: {length}"
            )

    return length


def merge_old_and_new(
    old_df: DataFrame,
    new_df: DataFrame,
    source: Union[str, None] = None,
    verbose: bool = True,
    key_cols: Union[str, List, None] = None,
) -> DataFrame:
    """
    Merge old and new DataFrames based on a set of key columns.

    This function compares two datasets (`old_df` and `new_df`)
    and performs an outer merge using
    specified key columns. It returns a DataFrame that includes:
    - Rows only in the old data
    - Rows in both old and new (with new values preferred)
    - Rows only in the new data

    Note
    ----
    This function is not intended to by explicitly executed by the user.

    Parameters
    ----------
    old_df : pd.DataFrame
        The original (historical) dataset.

    new_df : pd.DataFrame
        The newly downloaded or updated dataset to be merged.

    source : str or None, optional
        If `key_cols` is not provided, this value will be used to
        look up default key columns
        from `DEFAULT_KEY_COLS[source]`.

    verbose : bool, default=True
        If True, prints the number of rows in each merge category.

    key_cols : str, list, or None, optional
        Column name(s) used to join the old and new data. If None,
        `source` must be provided.

    Returns
    -------
    pd.DataFrame
        A merged DataFrame that contains all unique records from the
        old and new datasets, giving priority to values from the new
        dataset where overlap occurs.
    """
    assert isinstance(old_df, DataFrame), (
        "Expected old_df to be a pd.DataFrame, "
        + f"got {type(old_df)} instead."
    )
    assert isinstance(new_df, DataFrame), (
        "Expected old_df to be a pd.DataFrame, "
        + f"got {type(new_df)} instead."
    )

    # Check if empty
    if len(new_df) == 0:
        if verbose:
            print("No merge required.")
        return old_df

    # Get key cols
    if key_cols is None and source is None:
        raise ValueError("Can not be both 'source' and 'key_cols' None")
    elif key_cols is None:
        key_cols = DEFAULT_KEY_COLS[source]
    else:
        key_cols = parse_to_iter(key_cols)

    # Merge into 1
    merged = merge(
        old_df,
        new_df,
        on=key_cols,
        suffixes=("_old", "_new"),
        how="outer",
        indicator=True,
    )

    def clean(df, suffix="new", key_cols=None):
        """Return cleaned DataFrame with only one suffix + key cols."""
        cols = [c for c in df.columns if c.endswith(f"_{suffix}")]
        base = [c.replace(f"_{suffix}", "") for c in cols]

        result = df[cols].copy()
        result.columns = base

        # Reattach key columns from original (no suffix)
        if key_cols is not None:
            for key in key_cols:
                if key not in result.columns and key in df.columns:
                    result[key] = df[key]

        return result.reset_index(drop=True)

    # Get cleaned
    old_clean = clean(
        merged[merged["_merge"] == "left_only"],
        suffix="old",
        key_cols=key_cols,
    )
    upd_clean = clean(
        merged[merged["_merge"] == "both"], suffix="new", key_cols=key_cols
    )
    new_clean = clean(
        merged[merged["_merge"] == "right_only"],
        suffix="new",
        key_cols=key_cols,
    )

    # Message
    if verbose:
        nold = len(old_clean)
        nupd = len(upd_clean)
        nnew = len(new_clean)
        print(f" Rows in only old | both | only new: {nold} | {nupd} | {nnew}")

    # Concatenate
    latest = concat([old_clean, upd_clean, new_clean], ignore_index=True)

    # latest.drop_duplicates(inplace=True)
    latest = latest[old_df.columns]  # Reorder

    return latest



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# This file indicates that the directory should be treated as a package.

# ============================================================================
# DOCS
# ============================================================================

"""The ResoKit.utils.mass_radius package includes tools for M-R analysis."""

# =============================================================================
# IMPORTS
# =============================================================================

from .models import estimate_radius  # noqa
from .models import estimate_mass  # noqa

# Make the functions available at the package level

__all__ = [
    "estimate_radius",
    "estimate_mass",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# This file indicates that the directory should be treated as a package.

# ============================================================================
# DOCS
# ============================================================================
"""ResoKit.

ResoKit addresses the need for diagnosing and analyzing  mean motion
resonances (MMR) in coplanar planetary systems.
"""

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"


# =============================================================================
# IMPORTS
# =============================================================================

from . import core  # noqa
from . import datasets  # noqa
from . import load  # noqa
from . import query  # noqa
from . import utils  # noqa
from . import units  # noqa
from .utils import mmr  # noqa
from .utils import mass_radius  # noqa

# Make the core classes available directly from the package.

__all__ = [
    "core",
    "datasets",
    "load",
    "query",
    "units",
    "utils",
    "mmr",
    "mass_radius",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the
#   ResoKit Project (https://github.com/Gianuzzi/resokit).
# Copyright (c) 2025, Emmanuel Gianuzzi
# License: MIT
#   Full Text: https://github.com/Gianuzzi/resokit/blob/master/LICENSE

# This file indicates that the directory should be treated as a package.

# =============================================================================
# DOCS
# =============================================================================

"""The ResoKit.utils package includes tools for data analysis."""

# =============================================================================
# IMPORTS
# =============================================================================

from . import mass_radius  # noqa
from . import mmr  # noqa
from .utils import calc_a  # noqa
from .utils import calc_hill_radius  # noqa
from .utils import calc_period  # noqa
from .utils import float_to_fraction  # noqa

# Make the functions available at the package level

__all__ = [
    "mass_radius",
    "mmr",
    "calc_a",
    "calc_hill_radius",
    "calc_period",
    "float_to_fraction",
]
